"""Module with the core full reconstruction chain."""

from typing import List, Union

import numpy as np
import torch
import yaml
from torch_scatter import scatter_mean, scatter_std

from spine.data import IndexBatch, RunInfo, TensorBatch
from spine.utils.calib import CalibrationManager
from spine.utils.cluster.label import ClusterLabelAdapter
from spine.utils.ghost import ChargeRescaler
from spine.utils.globals import (
    CLUST_COL,
    COORD_COLS,
    DELTA_SHP,
    GHOST_SHP,
    MICHL_SHP,
    SHAPE_COL,
    SHOWR_SHP,
    TRACK_SHP,
    VALUE_COL,
)
from spine.utils.gnn.cluster import form_clusters_batch, get_cluster_label_batch
from spine.utils.gnn.evaluation import primary_assignment_batch
from spine.utils.logger import logger
from spine.utils.ppn import ParticlePointPredictor

from .graph_spice import GraphSPICE, GraphSPICELoss
from .grappa import GrapPA, GrapPALoss
from .layer.common.dbscan import DBSCAN
from .uresnet import SegmentationLoss, UResNetSegmentation
from .uresnet_ppn import UResNetPPN, UResNetPPNLoss

# from spine.model.layer.cnn.cnn_encoder import SparseResidualEncoder ??
# TODO: replace this with MultiParticleImageClassifier
# TODO: raname it something more generic like ParticleClusterImageClassifier?


class FullChain(torch.nn.Module):
    """Full reconstruction in all its glory.

    Modular, end-to-end particle imaging detector reconstruction chain:
    - Deghosting for 3D tomographic reconstruction artifiact removal
    - Voxel-wise semantic segmentation
    - Point proposal
    - Particle clustering
    - Shower primary identification
    - Interaction clustering
    - Particle type classification
    - Primary identification
    - Track orientation

    Typical configuration can look like this:

    .. code-block:: yaml

        model:
          name: grappa
          modules:
            chain:
               <dictionary of arguments to specify chain-wide configuration>
            uresnet_deghost:
               name: <name of the model used to deghost
               <dictionary of arguments to specify the deghosting module>
            uresnet_segmentation:
               name: <name of the model used to do segmentation>
               <dictionary of arguments to specify the segmentation module>
            dbscan:
               TODO
            graph_spice:
               name: <name of the model used to do segmentation>
               <dictionary of arguments to specify the segmentation module>
            grappa_shower:
               name: <name of the model used to do segmentation>
               <dictionary of arguments to specify the segmentation module>
            grappa_track:
               name: <name of the model used to do segmentation>
               <dictionary of arguments to specify the segmentation module>
            grappa_inter:
              TODO
            calibration:
              TODO

    See configuration file(s) prefixed with `full_chain_` under the `config`
    directory for detailed examples of working configurations.

    The `chain` section enables or disables specific stages of the full
    chain. When a module is disabled through this section, it will not be
    constructed. The configuration blocks for each enabled module should
    also live under the `modules` section of the configuration.
    """

    # TODO: update
    MODULES = [
        "grappa_shower",
        "grappa_track",
        "grappa_inter",
        "grappa_shower_loss",
        "grappa_track_loss",
        "grappa_inter_loss",
        "full_chain_loss",
        "graph_spice",
        "graph_spice_loss",
        "fragment_clustering",
        "chain",
        "dbscan_frag",
        ("uresnet_ppn", ["uresnet", "ppn"]),
    ]

    # Valid chain stage modes
    _modes = (
        ("deghosting", ("uresnet",)),
        ("charge_rescaling", ("collection", "average")),
        ("segmentation", ("uresnet",)),
        ("point_proposal", ("ppn",)),
        ("fragmentation", ("dbscan", "graph_spice", "dbscan_graph_spice")),
        ("shower_aggregation", ("skip", "grappa")),
        ("shower_primary", ("skip", "grappa")),
        ("track_aggregation", ("skip", "grappa")),
        ("particle_aggregation", ("skip", "grappa")),
        ("inter_aggregation", ("grappa",)),
        ("particle_identification", ("grappa", "image")),
        ("primary_identification", ("grappa",)),
        ("orientation_identification", ("grappa",)),
        ("calibration", ("apply",)),
    )

    # Valid calibration stages
    _calib_stages = (
        "segmentation",
        "fragmentation",
        "particle_aggregation",
        "inter_aggregation",
        "particle_identification",
    )

    def __init__(
        self,
        chain,
        uresnet_deghost=None,
        uresnet=None,
        uresnet_ppn=None,
        adapt_labels=None,
        predict_points=None,
        graph_spice=None,
        dbscan=None,
        grappa_shower=None,
        grappa_track=None,
        grappa_particle=None,
        grappa_inter=None,
        calibration=None,
        uresnet_deghost_loss=None,
        uresnet_loss=None,
        uresnet_ppn_loss=None,
        graph_spice_loss=None,
        grappa_shower_loss=None,
        grappa_track_loss=None,
        grappa_particle_loss=None,
        grappa_inter_loss=None,
    ):
        """Initialize the full chain model.

        Parameters
        ----------
        chain : dict
            Dictionary of parameters used to configure the chain
        uresnet_deghost : dict, optional
            Deghosting model configuration
        uresnet_ppn : dict, optional
            Segmentation and point proposal model configuration
        adapt_labels : dict, optional
            Parameters for the cluster label adaptation (if non-standard)
        predict_points : dict, optional
            Parameters for the particle point predictor (if non-standard)
        dbscan : dict, optional
            Connected component clustering configuration
        graph_spice : dict, optional
            Supervised connected component clustering model configuration
        grappa_shower : dict, optional
            Shower aggregation model configuration
        grappa_track : dict, optional
            Track aggregation model configuration
        grappa_particle : dict, optional
            Global particle aggregation configuration
        grappa_inter : dict, optional
            Interaction aggregation model configuration
        calibration : dict, optional
            Calibration manager configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the main chain configureation
        process_chain_config(self, **chain, dump_config=True)

        # Initialize the deghosting model
        if self.deghosting is not None and self.deghosting == "uresnet":
            assert uresnet_deghost is not None, (
                "If the deghosting is using UResNet, must provide the "
                "`uresnet_deghost` configuration block."
            )
            self.uresnet_deghost = UResNetSegmentation(uresnet_deghost)

        # Initialize the charge rescaling process (adapt to ghost predictions)
        if self.charge_rescaling is not None:
            self.charge_rescaler = ChargeRescaler(
                collection_only=self.charge_rescaling == "collection"
            )

        # Initialize the semantic segmentation model (+ point proposal)
        if self.segmentation is not None and self.segmentation == "uresnet":
            assert (uresnet is not None) ^ (uresnet_ppn is not None), (
                "If the segmentation is using UResNet, must provide the "
                "`uresnet` or `uresnet_ppn` configuration block."
            )
            if uresnet is not None:
                self.uresnet = UResNetSegmentation(uresnet)
            else:
                self.uresnet_ppn = UResNetPPN(**uresnet_ppn)

        # Initialize the relabeling process (adapt to the semantic predictions)
        self.label_adapter = ClusterLabelAdapter(**(adapt_labels or {}))

        # Initialize the point predictor (for fragment/particle clusters)
        self.point_predictor = ParticlePointPredictor(**(predict_points or {}))

        # Initialize the dense clustering model
        self.fragment_shapes = []
        if self.fragmentation is not None and (
            "dbscan" in self.fragmentation or "graph_spice" in self.fragmentation
        ):
            if "dbscan" in self.fragmentation:
                assert dbscan is not None, (
                    "If the fragmentation is done using DBSCAN, must "
                    "provide the `dbscan` configuration block."
                )
                self.dbscan = DBSCAN(**dbscan)
                self.fragment_shapes.extend(self.dbscan.shapes)
            if "graph_spice" in self.fragmentation:
                assert graph_spice is not None, (
                    "If the fragmentation is done using Graph-SPICE, must "
                    "provide the `graph_spice` configuration block."
                )
                self.graph_spice = GraphSPICE(graph_spice)
                self.fragment_shapes.extend(self.graph_spice.shapes)

            # Check that there is no redundancy between the fragmenters
            uniques, counts = np.unique(self.fragment_shapes, return_counts=True)
            assert np.all(uniques == np.arange(4)), (
                "All four expected semantic classes should be fragmented "
                "by either DBSCAN or Graph-SPICE."
            )
            assert np.all(counts) == 1, (
                "Some of the shapes appear in both the DBSCAN and the "
                "Graph-SPICE based fragmentation. Ambiguous."
            )

        # Initialize the graph-based aggregator modules
        grappa_configs = {
            "shower": grappa_shower,
            "track": grappa_track,
            "particle": grappa_particle,
            "inter": grappa_inter,
        }
        for stage, config in grappa_configs.items():
            if getattr(self, f"{stage}_aggregation") == "grappa":
                name = f"grappa_{stage}"
                assert config is not None, (
                    f"If the {stage} aggregation is done using GrapPA, "
                    f"must provide the {name} configuration block."
                )
                setattr(self, name, GrapPA(config))
                assert (
                    getattr(self, name).make_groups == True
                ), "The aggregators should have `make_groups: true`"

        # Initialize the standalone particle classifier
        # TODO (likely an image classifier)

        # Initialize the interaction-classification module
        # TODO (could be done by either CNN or graph-level GNN)

        # Initialize the calibrator manager
        self.calibration_stage = None
        if self.calibration == "apply":
            assert calibration is not None, (
                "If the calibration is to be applied, must provide the "
                "`calibration` configuration block."
            )
            assert "stage" in calibration, (
                "If the calibration is to be applied, must provide the "
                "`stage` to specify where to apply it."
            )
            self.calibration_stage = calibration.pop("stage")
            self.calibrator = CalibrationManager(**calibration)
            calibration["stage"] = self.calibration_stage

    @property
    def modes(self):
        """Dictionary of (stage, modes) pairs which determine which options
        are available to each of the the reconstruction stage.

        Returns
        -------
        Dict[str, Tuple(str)]
            Dictionary of (stage, modes) pairs to be used
        """
        return dict(self._modes)

    def forward(
        self,
        data,
        sources=None,
        seg_label=None,
        clust_label=None,
        coord_label=None,
        energy_label=None,
        meta=None,
        run_info=None,
    ):
        """Run a batch of data through the full chain.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is the number of features per voxel
        sources : TensorBatch, optional
            (N, 2) tensor of module/tpc pair for each voxel
        seg_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of segmentation labels
            - 1 is the segmentation label
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels
        coord_label : TensorBatch, optional
            (N, 1 + D + N_p) Tensor of point of interest labels
            - N_p is the number point labels
        energy_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of true energy deposition values
            - 1 is the energy deposition value in each voxel
        meta : Meta, optional
            Image metadata information
        run_info : List[RunInfo], optional
            Object containing information about the run, subrun and event

        Returns
        -------
        TODO
        """
        # Initialize the full chain output dictionary
        self.result = {}

        # Run the deghosting step
        data, sources = self.run_deghosting(data, sources, seg_label, clust_label)

        # Run the semantic segmentation (and point proposal) stage
        if self.calibration_stage == "segmentation":
            data = self.run_calibration(data, sources, energy_label, meta, run_info)
        clust_label = self.run_segmentation_ppn(data, seg_label, clust_label)

        # Run the fragmentation stage
        if self.calibration_stage == "fragmentation":
            data = self.run_calibration(data, sources, energy_label, meta, run_info)
        self.run_fragmentation(data, clust_label)

        # Run the particle aggregation
        if self.calibration_stage == "particle_aggregation":
            data = self.run_calibration(data, sources, energy_label, meta, run_info)
        self.run_part_aggregation(data, clust_label, coord_label)

        # Run the interaction aggregation
        if self.calibration_stage == "inter_aggregation":
            data = self.run_calibration(data, sources, energy_label, meta, run_info)
        self.run_inter_aggregation(data, clust_label, coord_label)

        # Run an independant particle classification stage
        if self.calibration_stage == "particle_classification":
            data = self.run_calibration(data, sources, energy_label, meta, run_info)
        # TODO

        # Run the interaction classification
        # TODO

        # Return
        return self.result

    def run_deghosting(self, data, sources=None, seg_label=None, clust_label=None):
        """Run the deghosting algorithm.

        This removes points that are artifacts of the tomographic
        reconstruction. This is only relevant for detectors producing 2D
        projections of an event. If requested, charge is rescaled according
        to the deghosting mask.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        sources : TensorBatch, optional
            (N, 2) tensor of module/tpc pair for each voxel
        seg_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of segmentation labels
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels

        Returns
        -------
        TensorBatch
            (N, 1 + D + N_f) tensor of deghosted voxel/value pairs
        """
        if self.deghosting == "uresnet":
            # Pass the data through the model
            res_deghost = self.uresnet_deghost(data)

            # Store the ghost scores and the ghost mask
            ghost_tensor = res_deghost["segmentation"].tensor
            ghost_pred = torch.argmax(ghost_tensor, dim=1)
            data_adapt = TensorBatch(
                data.tensor[ghost_pred == 0],
                batch_size=data.batch_size,
                has_batch_col=True,
                coord_cols=data.coord_cols,
            )
            ghost_pred = TensorBatch(ghost_pred, data.counts)

            # Rescale the charge, if requested
            if self.charge_rescaling is not None:
                charges = self.charge_rescaler(data_adapt)
                tensor_deghost = data_adapt.tensor[:, :-6]
                tensor_deghost[:, VALUE_COL] = charges
                data_adapt.data = tensor_deghost

            self.result["ghost"] = res_deghost["segmentation"]
            self.result["ghost_pred"] = ghost_pred
            self.result["data_adapt"] = data_adapt

            # If sources are provided, narrow them down to non-ghosts
            sources_adapt = None
            if sources is not None:
                # Store the sources of the predicted non-ghosts
                sources_adapt = TensorBatch(
                    sources.tensor[ghost_pred.tensor == 0], data_adapt.counts
                )
                self.result["sources_adapt"] = sources_adapt

                # Store the sources of the true non-ghosts
                if seg_label is not None:
                    ghost_label = seg_label.tensor[:, SHAPE_COL] < GHOST_SHP
                    if clust_label is not None:
                        counts = clust_label.counts
                    else:
                        counts = TensorBatch(
                            seg_label.tensor[ghost_label],
                            batch_size=seg_label.batch_size,
                            has_batch_col=True,
                        ).counts

                    sources_label = TensorBatch(sources.tensor[ghost_label], counts)
                    self.result["sources_label"] = sources_label

            return data_adapt, sources_adapt

        elif self.deghosting == "label":
            # Use ghost labels to remove ghost voxels from the input
            assert seg_label is not None, "Must provide `seg_label` to deghost with it."
            ghost_pred = (seg_label.tensor[:, SHAPE_COL] == GHOST_SHP).long
            tensor_deghost = data.tensor[ghost_pred == 0]

            # Use the label rescaled charge, if requested
            if self.charge_rescaling is not None:
                assert clust_label is not None, (
                    "Must provide `clust_label` to fetch the true " "rescaled charge."
                )
                tensor_deghost[:, VALUE_COL] = clust_label[:, VALUE_COL]

            # Store and return
            ghost_pred = TensorBatch(ghost_pred, data.counts)
            data_adapt = TensorBatch(
                tensor_deghost, batch_size=data.batch_size, coord_cols=data.coord_cols
            )
            self.result["ghost_pred"] = ghost_pred
            self.result["data_adapt"] = data_adapt

            # If sources are provided, narrow them down to non-ghosts
            sources_adapt = None
            if sources is not None:
                sources_adapt = TensorBatch(
                    sources.tensor[ghost_pred.tensor == 0], data_adapt.counts
                )
                self.result["sources_adapt"] = sources_adapt
                if seg_label is not None:
                    self.result["sources_label"] = sources_adapt

            return data_adapt, sources_adapt

        else:
            # If sources are provided, set the expected output
            sources_adapt = None
            if sources is not None:
                self.result["sources_adapt"] = sources
                if seg_label is not None:
                    self.result["sources_label"] = sources

            # Nothing to do
            return data, sources_adapt

    def run_segmentation_ppn(self, data, seg_label=None, clust_label=None):
        """Run the semantic segmentation and the point proposal algorithms.

        This classifies each individual voxel in the image into different
        particle topological categories and identifies poins of interest,
        namely track end points and shower fragment start points.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        seg_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of segmentation labels
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
        """
        if self.segmentation == "uresnet":
            # Run the data through the appropriate model
            if hasattr(self, "uresnet"):
                res_seg = self.uresnet(data)
            else:
                res_seg = self.uresnet_ppn(data)

            # If the deghosting is done as part of this step, process it
            if "ghost" in res_seg:
                # Store the ghost scores and the ghost mask
                ghost_tensor = res_deghost["host"].tensor
                ghost_pred = torch.argmax(ghost_tensor, dim=1)
                data_adapt = TensorBatch(
                    data.tensor[ghost_pred == 0],
                    batch_size=data.batch_size,
                    coord_cols=data.coord_cols,
                )
                ghost_pred = TensorBatch(ghost_pred, data.counts)

                self.result["ghost_pred"] = ghost_pred
                self.result["data_adapt"] = data_adapt

                # If there are PPN outputs, deghost them
                if "ppn_points" in res_seg:
                    deghost_index = torch.where(ghost_pred == 0)[0]
                    res_seg["ppn_points"] = res_seg["ppn_points"][deghost_index]
                    for key in [
                        "ppn_masks",
                        "ppn_coords",
                        "ppn_layers",
                        "ppn_classify_endpoints",
                    ]:
                        if key in res_seg:
                            res_seg[key][-1] = res_seg[key][-1][deghost_index]

            # Update the result dictionary
            self.result.update(res_seg)
            seg_pred = torch.argmax(res_seg["segmentation"].tensor, dim=1)

            self.result["seg_pred"] = TensorBatch(seg_pred, data.counts)

            # If the rest of the chain is run, must adapt cluster labels now
            if seg_label is not None and clust_label is not None:
                seg_pred = self.result["seg_pred"]
                ghost_pred = self.result.get("ghost_pred", None)
                clust_label = self.label_adapter(
                    clust_label, seg_label, seg_pred, ghost_pred
                )

                self.result["clust_label_adapt"] = clust_label

        elif self.segmentation == "label":
            # Use the segmentation labels
            assert seg_label is not None, "Must provide `seg_label` to segment with it."
            seg_pred = seg_label.tensor[:, SHAPE_COL]

            self.result["seg_pred"] = TensorBatch(seg_pred, data.counts)

        return clust_label

    def run_fragmentation(self, data, clust_label=None):
        """Run the fragmentation algorithm, i.e. the dense clustering step.

        This breaks down each topological class individually into a set of
        fragments, each composed of contiguous voxels which belong to a single
        particle instance, but do not necessarily make up the whole instance.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
        """
        # Nothing to do if there is no fragmentation requested
        if self.fragmentation is None:
            return

        # Initialize the fragment-level output
        counts = np.zeros(data.batch_size, dtype=np.int64)
        fragments = IndexBatch([], data.edges[:-1], counts, [])
        fragment_shapes = TensorBatch(np.empty(0, dtype=np.int64), counts)

        # Append the fragment list
        if "dbscan" in self.fragmentation:
            # Run DBSCAN
            clusts, clust_shapes = self.dbscan(data, **self.result)

            # Append
            fragments = fragments.merge(clusts)
            fragment_shapes = fragment_shapes.merge(clust_shapes)

        if "graph_spice" in self.fragmentation:
            # Run Graph-SPICE
            seg_pred = TensorBatch(self.result["seg_pred"].tensor[:, None], data.counts)
            res_gs = self.graph_spice(data, seg_pred, clust_label)

            # Update the global result with the graph_spice output
            self.result.update(
                {f"graph_spice_{key}": val for key, val in res_gs.items()}
            )

            # If fragments are produced, store them
            if "clusts" in res_gs:
                # Fragments point at a partial tensor, adapt them
                clusts = res_gs["clusts"]
                clust_shapes = res_gs["clust_shapes"]
                filter_index = res_gs["filter_index"].index
                for i, f in enumerate(clusts.index_list):
                    clusts.data[i] = filter_index[f]

                clusts.offsets = data.edges[:-1]

                # Append
                fragments = fragments.merge(clusts.to_numpy())
                fragment_shapes = fragment_shapes.merge(clust_shapes)

        if self.fragmentation == "label":
            assert (
                clust_label is not None
            ), "Must provide `clust_label` to use it for fragmentation."
            fragments = form_clusters_batch(clust_label.to_numpy(), column=CLUST_COL)
            fragment_shapes = get_cluster_label_batch(clust_label, fragments)

        if fragments is not None:
            self.result["fragment_clusts"] = fragments
            self.result["fragment_shapes"] = fragment_shapes

    def run_part_aggregation(self, data, clust_label=None, coord_label=None):
        """Run the particle aggreation step.

        This step gathers particle fragments into complete particle instances.
        It either aggregates shower and track fragments independently or
        jointly into a single step.

        In the process of shower aggregation, shower primaries can
        be identified.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
        coord_label : TensorBatch, optional
            (N, 1 + D + 6) Array of label particle end points
        """
        # Fetch the list of fragments and semantic classes
        if "fragment_clusts" not in self.result:
            return

        fragments = self.result["fragment_clusts"]
        fragment_shapes = self.result["fragment_shapes"]

        # If GrapPA is run on track/shower separately, prepare merged products
        merge = {}
        if self.shower_aggregation == "grappa" or self.track_aggregation == "grappa":
            num_clusts = len(fragments.index_list)
            kwargs = {"dtype": data.dtype, "device": data.device}
            merge["start_points"] = torch.full((num_clusts, 3), -np.inf, **kwargs)
            merge["end_points"] = torch.full((num_clusts, 3), -np.inf, **kwargs)
            merge["node_pred"] = torch.full((num_clusts, 2), -np.inf, **kwargs)
            merge["group_pred"] = -np.ones(num_clusts, dtype=np.int64)

        # Initialize the particle-level output
        counts = np.zeros(fragments.batch_size, dtype=np.int64)
        particles = IndexBatch([], fragments.offsets, counts, [])
        particle_shapes = TensorBatch(np.empty(0, dtype=np.int64), counts)
        particle_primaries = IndexBatch([], fragments.offsets, counts, [])

        # Loop over GraPA models, append the particle list
        shapes = {
            "shower": [SHOWR_SHP, MICHL_SHP, DELTA_SHP],
            "track": [TRACK_SHP],
            "particle": self.fragment_shapes,
        }
        use_primary = {"shower": True, "track": False, "particle": True}
        for name in ["shower", "track", "particle"]:
            # Dispatch the input to the right aggregation method
            switch = getattr(self, f"{name}_aggregation")

            if switch == "grappa":
                # Use GraPA to aggregate instances
                prefix = f"{name}_fragment" if name != "particle" else "fragment"
                model = getattr(self, f"grappa_{name}")
                groups, group_shapes, group_primaries, shape_index = self.run_grappa(
                    prefix,
                    model,
                    data,
                    fragments,
                    fragment_shapes,
                    coord_label,
                    aggregate_shapes=True,
                    shape_use_primary=use_primary[name],
                    retain_primaries=use_primary[name],
                )

                if merge and shape_index is not None:
                    for key in merge:
                        if key != "group_pred":
                            if key != "node_pred" or use_primary[name]:
                                merge[key][shape_index] = self.result[
                                    f"{prefix}_{key}"
                                ].tensor
                        else:
                            max_id = np.max(merge[key]) + 1
                            merge[key][shape_index] = (
                                self.result[f"{prefix}_{key}"].tensor + max_id
                            )

            elif switch == "label":
                # Use cluster labels to aggregate instances
                groups, group_shapes, group_primaries, shape_index = self.group_labels(
                    shapes[name],
                    data,
                    fragments,
                    fragment_shapes,
                    aggregate_shapes=True,
                    shape_use_primary=use_primary[name],
                    retain_primaries=use_primary[name],
                )

            elif switch == "skip":
                # Leave the shower fragments as is
                groups, group_shapes, shape_index = self.restrict_clusts(
                    fragments, fragment_shapes, shapes[name]
                )
                group_primaries = groups

            else:
                # Skip if there nothing to do
                continue

            # Append
            particles = particles.merge(groups)
            particle_shapes = particle_shapes.merge(group_shapes)
            particle_primaries = particle_primaries.merge(group_primaries)

        # Store merged fragment properties, if needed
        if merge:
            for key, value in merge.items():
                # Convert to a TensorBatch
                coord_cols = np.array([0, 1, 2]) if key.endswith("points") else None
                self.result[f"fragment_{key}"] = TensorBatch(
                    value, counts=fragments.counts, coord_cols=coord_cols
                )

                # For group predictions, reorganize them by batch entry
                if key == "group_pred":
                    group_pred = self.result[f"fragment_{key}"]
                    offset = 0
                    for b, group_pred_b in enumerate(group_pred.split()):
                        lower, upper = group_pred.edges[b], group_pred.edges[b + 1]
                        group_pred_b = np.unique(group_pred_b, return_inverse=True)[-1]
                        group_pred.data[lower:upper] = offset + group_pred_b
                        offset += len(group_pred_b)

        # Store particle objects
        self.result["particle_clusts"] = particles
        self.result["particle_shapes"] = particle_shapes
        self.result["particle_primaries"] = particle_primaries

    def run_inter_aggregation(self, data, clust_label=None, coord_label=None):
        """Run the interaction aggreation step.

        This step gathers particles into complete interaction instances.

        In the process of interaction aggregation, other tasks may be performed:
        - Particle identification
        - Primary tagging (particle coming from the interaction vertex)
        - Orientation tagging (order start/end points of particles)
        - Vertex reconstruction

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
        coord_label : TensorBatch, optional
            (N, 1 + D + 6) Array of label particle end points
        """
        # Fetch the list of particles and semantic classes
        if "particle_clusts" not in self.result:
            return

        particles = self.result["particle_clusts"]
        particle_shapes = self.result["particle_shapes"]
        particle_primaries = self.result["particle_primaries"]

        # Append the interaction list
        if self.inter_aggregation == "grappa":
            # Use GraPA to aggregate instances
            interactions, _, _, _ = self.run_grappa(
                "particle",
                self.grappa_inter,
                data,
                particles,
                particle_shapes,
                particle_primaries,
                coord_label,
                point_use_primary=True,
            )

        elif self.inter_aggregation == "label":
            # Use cluster labels to aggregate instances
            interactions, _, _, _ = self.group_labels(
                shapes[name], data, particles, particle_shapes
            )

        # Store interaction objects
        if self.inter_aggregation is not None:
            self.result["interaction_clusts"] = interactions

    def run_calibration(
        self, data, sources=None, energy_label=None, meta=None, run_info=None
    ):
        """Run the calibration algorithm.

        This converts the raw charge values in ADC to energy depositions
        expressed in MeV. It applies gain, recombination, transparency
        and electron lifetime corrections.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        sources : TensorBatch, optional
            (N, 2) tensor of module/tpc pair for each voxel
        energy_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of true energy deposition values
            - 1 is the energy deposition value in each voxel
        meta : Meta, optional
            Image metadata information
        run_info : List[RunInfo], optional
            Object containing information about the run, subrun and event

        Returns
        -------
        TensorBatch
            (N, 1 + D + N_f) tensor of calibrated voxel/value pairs
        """
        if self.calibration == "apply":
            # Check that the metadata is provided
            assert meta is not None, (
                "Must provide the metadata to convert pixel coordinates "
                "to detector coordinates and apply calibrations."
            )

            # Loop over entries in the batch (might have different meta/run IDs)
            data_np = data.to_numpy().tensor
            sources = sources.to_numpy().tensor if sources is not None else None
            rep = data.batch_size // len(meta)
            for b in range(data.batch_size):
                # Fetch necessary information for this batch entry
                lower, upper = data.edges[b], data.edges[b + 1]
                data_b = data_np[lower:upper]
                voxels_b = data_b[:, COORD_COLS]
                values_b = data_b[:, VALUE_COL]
                sources_b = sources[lower:upper] if sources is not None else None

                # Fetch meta/run information for this batch entry
                meta_b = meta[b // rep]
                run_id = run_info[b // rep].run if run_info is not None else None

                # Calibrate voxel values
                values_b = self.calibrator(
                    voxels_b,
                    values_b,
                    sources_b,
                    run_id,
                    meta=meta_b,
                    module_id=b % rep,
                )

                data.tensor[lower:upper, VALUE_COL] = torch.tensor(
                    values_b, dtype=data.dtype, device=data.device
                )

            self.result["data_adapt"] = data

        elif self.calibration == "label":
            # Use energy labels to give values to each voxel
            assert (
                energy_label is not None
            ), "Must provide `energy_label` to do label-based calibration."
            data.tensor[:, VALUE_COL] = energy_label.tensor[:, VALUE_COL]

            self.result["data_adapt"] = data

        return data

    def run_grappa(
        self,
        prefix,
        model,
        data,
        clusts,
        clust_shapes,
        clust_primaries=None,
        coord_label=None,
        aggregate_shapes=False,
        shape_use_primary=False,
        point_use_primary=False,
        retain_primaries=False,
    ):
        """Run the GNN-based particle aggregator.

        Parameters
        ----------
        prefix : str
            Name of the aggregation step
        model : GraPA
            GraPA model to execute for this aggregation step
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clusts : IndexBatch
            List of clusters to aggregate using GrapPA
        clust_shapes : TensorBatch
            Semantic type of each of the clusters
        clust_primaries : IndexBatch
            List of primary fragments associated with each input cluster
        coord_label : TensorBatch, optional
            (N, 1 + D + 6) Array of label particle end points
        aggregate_shapes : bool, default False
            Combine shapes to give a shape to the aggregated object
        shape_use_primary : bool, default False
            Use primary shape as the group shape
        point_use_primary : bool, default False
            Use the primary fragment to get the points
        retain_primaries : bool, default False
            Retain the primary cluster in the aggregated group

        Returns
        -------
        groups : IndexBatch
            List of cluster groups aggregated using GrapPA
        group_shapes : TensorBatch
            Semantic type of each of the cluster groups
        group_primaries : IndexBatch
            List of primary clusters for each group
        shape_index : np.ndarray
            List of indexes used to restrict the original cluster list
        """
        # Restrict the clusters to those in the input of the model
        clusts, clust_shapes, shape_index = self.restrict_clusts(
            clusts, clust_shapes, model.node_type
        )

        # Prepare the input to the aggregation stage
        grappa_input = self.prepare_grappa_input(
            model,
            data,
            clusts,
            clust_shapes,
            clust_primaries,
            coord_label,
            point_use_primary,
        )

        # Pass it through GrapPA, produce shower instances
        res_grappa = model(**grappa_input)
        self.result.update({f"{prefix}_{k}": v for k, v in res_grappa.items()})

        # If requested, convert the node predictions to a primary mask
        group_pred = res_grappa["group_pred"]
        primary_mask = None
        if shape_use_primary or retain_primaries:
            assert "node_pred" in res_grappa, (
                "Must provide node predictions to use primary shape "
                "or preserve the list of primary clusters."
            )
            node_pred = res_grappa["node_pred"].to_numpy()
            primary_mask = primary_assignment_batch(node_pred, group_pred)

        # Build shower instances, get their semantic type
        return (
            *self.build_groups(
                clusts,
                clust_shapes,
                group_pred,
                primary_mask,
                aggregate_shapes,
                shape_use_primary,
                retain_primaries,
            ),
            shape_index,
        )

    def group_labels(
        self,
        data,
        clusts,
        clust_shapes,
        aggregate_shapes=False,
        shape_use_primary=False,
        retain_primaries=False,
    ):
        """Aggregate particles using labels.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clusts : IndexBatch
            List of clusters to aggregate using GrapPA
        clust_shapes : TensorBatch
            Semantic type of each of the clusters
        aggregate_shapes : bool, default False
            Combine shapes to give a shape to the aggregated object
        shape_use_primary : bool, default False
            Use primary shape as the group shape
        retain_primaries : bool, default False
            Retain the primary cluster

        Returns
        -------
        groups : IndexBatch
            List of cluster groups aggregated using labels
        group_shapes : TensorBatch
            Semantic type of each of the cluster groups
        group_primaries : IndexBatch
            List of primary clusters for each group
        shape_index : np.ndarray
            List of indexes used to restrict the original cluster list
        """
        # Restrict the clusters to those in the input of the model
        clusts, clust_shapes, shape_index = self.restrict_clusts(
            clusts, clust_shapes, model.node_type
        )

        # If requested, convert the node predictions to a primary mask
        group_ids = get_cluster_label_batch(data, clusts, GROUP_COL)
        primary_mask = None
        if shape_use_primary:
            primary_mask = get_cluster_label_batch(data, clusts, PRGRP_COL)
            primary_mask = primary_mask.astype(bool)

        # Build shower instances, get their semantic type
        return (
            *self.build_groups(
                clusts,
                clust_shapes,
                group_ids,
                primary_mask,
                aggregate_shapes,
                shape_use_primary,
                retain_primaries,
            ),
            shape_index,
        )

    def restrict_clusts(self, clusts, clust_shapes, shapes):
        """Restricts a cluster list against a list of shapes.

        Parameters
        ----------
        clusts : IndexBatch
            List of clusters to aggregate using GrapPA
        clust_shapes : TensorBatch
            Semantic type of each of the clusters
        shapes : List[int]
            List of semantic shapes to restrict to

        Returns
        -------
        clusts : IndexBatch
            Restricted list of clusters
        clust_shapes : TensorBatch
            Restricted list of semantic types
        shape_index : np.ndarray
            List of indexes used to restrict the original cluster list
        """
        # Restrict the clusters to those in the input of the model
        shape_index = False
        if shapes != self.fragment_shapes:
            mask = np.zeros(len(clust_shapes.tensor), dtype=bool)
            for c in shapes:
                mask |= clust_shapes.tensor == c
            shape_index = np.where(mask)[0]

            batch_ids = clusts.batch_ids[shape_index]
            clusts_np = np.empty(len(clusts.index_list), dtype=object)
            clusts_np[:] = clusts.index_list
            clusts = IndexBatch(
                clusts_np[shape_index],
                offsets=clusts.offsets,
                single_counts=clusts.single_counts[shape_index],
                batch_ids=batch_ids,
                batch_size=clusts.batch_size,
            )
            clust_shapes = TensorBatch(clust_shapes.tensor[shape_index], clusts.counts)

        return clusts, clust_shapes, shape_index

    def prepare_grappa_input(
        self,
        model,
        data,
        clusts,
        clust_shapes,
        clust_primaries=None,
        coord_label=None,
        point_use_primaries=False,
    ):
        """Prepares the input to a GrapPA model.

        It builds the following input to GrpPA:
        - points: end points of fragments/particles
        - value: mean/std of charge distribution in each cluster
        - shape: shape of each fragment/particle

        Parameters
        ----------
        model : torch.nn.Module
            GrapPA model to feed information to
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clusts : IndexBatch
            List of clusters to aggregate using GrapPA
        clust_shapes : TensorBatch
            Semantic type of each of the clusters
        clust_primaries : IndexBatch, optional
            List of primary fragment within each cluster to aggregate
        coord_label : TensorBatch, optional
            (N, 1 + D + 6) Array of label particle end points
        point_use_primaries:
            Use the primary fragment only to infer primaries

        Returns
        -------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
        clusts : IndexBatch
            Input clusters to the model
        shapes : TensorBatch
            List of semantic type of each clusters
        points : TensorBatch
            List of start/end points associated with each cluster
        extra : TensorBatch
            List of additional features to pass to the GrapPA model
        """
        # Store the basics
        grappa_input = {}
        grappa_input["data"] = data
        grappa_input["clusts"] = clusts
        grappa_input["shapes"] = clust_shapes
        if coord_label is not None:
            grappa_input["coord_label"] = coord_label

        # Get the particle end points, if requested
        if hasattr(model.node_encoder, "add_points") and model.node_encoder.add_points:
            # Fetch the cluster list to use to get points
            ref_clusts = clusts
            if point_use_primaries:
                assert (
                    clust_primaries is not None
                ), "If using primaries to get points, must provide them."
                ref_clusts = clust_primaries

            # Get and store the points
            points = self.point_predictor(
                data, ref_clusts, clust_shapes, self.result["ppn_points"]
            )

            grappa_input["points"] = points

        # Get the supplemental information, if requested
        if hasattr(model.node_encoder, "add_value") and (
            model.node_encoder.add_value or model.node_envoder.add_shape
        ):
            extra = []
            if model.node_encoder.add_value:
                values = data.tensor[clusts.full_index, VALUE_COL]
                index_ids = torch.tensor(
                    clusts.index_ids, dtype=torch.long, device=data.device
                )
                extra.append(scatter_mean(values, index_ids))
                extra.append(scatter_std(values, index_ids))

            if model.node_encoder.add_shape:
                shapes = torch.tensor(
                    clust_shapes.tensor, dtype=data.dtype, device=data.device
                )
                extra.append(shapes)

            grappa_input["extra"] = TensorBatch(torch.stack(extra).t(), clusts.counts)

        return grappa_input

    def build_groups(
        self,
        clusts,
        clust_shapes,
        group_pred,
        primary_mask=None,
        aggregate_shapes=False,
        shape_use_primary=False,
        retain_primaries=False,
    ):
        """Use groups predictions from GrapPA to build superstructures.

        Parameters
        ----------
        clusts : IndexBatch
            List of clusters to aggregate using GrapPA
        clust_shapes : TensorBatch
            Semantic type of each of the clusters
        group_pred : TensorBatch
            Group ID of each node in the GraPA output
        primary_mask : TensorBatch
            Binary mask as to whether a node is a group primary or not
        aggregate_shapes : bool, default False
            Combine shapes to give a shape to the aggregated object
        shape_use_primary : bool, default False
            Use primary shape as the group shape
        retain_primaries : bool, default False
            Retain the primary cluster in the aggregated group

        Returns
        -------
        groups : IndexBatch
            List of cluster groups aggregated using GrapPA
        group_shapes : TensorBatch
            Semantic type of each of the cluster groups
        group_primaries : IndexBatch
            List of primary clusters for each group
        """
        # Cast node_pred to numpy if it is provided
        groups, group_shapes, group_primaries = [], [], []
        counts, single_counts, single_primary_counts = [], [], []
        for b in range(group_pred.batch_size):
            # Fetch the subset of data for this batch
            clusts_b = clusts[b]
            offset_b = clusts.offsets[b]
            group_pred_b = group_pred[b]
            if aggregate_shapes:
                clust_shapes_b = clust_shapes[b]
            if primary_mask is not None:
                primary_mask_b = primary_mask[b]

            # Loop over unique group IDs
            group_ids = np.unique(group_pred_b)
            counts.append(len(group_ids))
            for g in group_ids:
                # Build the set of groups made up of input clusters
                group_index = np.where(group_pred_b == g)[0]
                groups.append(offset_b + np.concatenate(clusts_b[group_index]))
                single_counts.append(len(groups[-1]))

                # Extract the shape and primary ID for this group
                if primary_mask is not None:
                    primary_id = group_index[primary_mask_b[group_index]][0]
                    if aggregate_shapes:
                        group_shapes.append(clust_shapes_b[primary_id])
                    if retain_primaries:
                        group_primaries.append(offset_b + clusts_b[primary_id])
                        single_primary_counts.append(len(group_primaries[-1]))

                elif aggregate_shapes:
                    shapes, shape_counts = np.unique(
                        clust_shapes_b[group_index], return_counts=True
                    )
                    group_shapes.append(shapes[np.argmax(shape_counts)])

        groups = IndexBatch(groups, clusts.offsets, counts, single_counts)
        if aggregate_shapes:
            group_shapes = np.array(group_shapes, dtype=np.int64)
            group_shapes = TensorBatch(group_shapes, counts)
        if retain_primaries:
            group_primaries = IndexBatch(
                group_primaries, clusts.offsets, counts, single_primary_counts
            )
        else:
            group_primaries = groups

        return groups, group_shapes, group_primaries


class FullChainLoss(torch.nn.Module):
    """Loss function for the full chain.

    See Also
    --------
    FullChain
    """

    # Valid chain stage modes
    _modes = FullChain._modes

    def __init__(
        self,
        chain,
        uresnet_deghost=None,
        uresnet_deghost_loss=None,
        uresnet=None,
        uresnet_loss=None,
        uresnet_ppn=None,
        uresnet_ppn_loss=None,
        graph_spice=None,
        graph_spice_loss=None,
        grappa_shower_loss=None,
        grappa_track_loss=None,
        grappa_particle_loss=None,
        grappa_inter_loss=None,
        **kwargs,
    ):
        """Initialize the full chain model.

        Parameters
        ----------
        chain : dict
            Dictionary of parameters used to configure the chain
        uresnet_deghost : dict, optional
            Deghosting model configuration
        uresnet_deghost_loss : dict, optional
            Deghosting loss configuration
        uresnet : dict, optional
            Segmentation model configuration
        uresnet_loss : dict, optional
            Segmentation loss configuration
        uresnet_ppn: dict, optional
            Segmentation and point proposal model configuration
        uresnet_ppn_loss : dict, optional
            Segmentation and point proposal loss configuration
        graph_spice : dict, optional
            Supervised connected component clustering model configuration
        graph_spice_loss : dict, optional
            Supervised connected component clustering loss configuration
        grappa_shower : dict, optional
            Shower aggregation model configuration
        grappa_track : dict, optional
            Track aggregation model configuration
        grappa_particle : dict, optional
            Global particle aggregation configuration
        grappa_inter : dict, optional
            Interaction aggregation model configuration
        """
        # Initialize the parent class
        super().__init__()

        # Process the main chain configureation
        process_chain_config(self, **chain)

        # Initialize the deghosting loss
        if self.deghosting == "uresnet" and uresnet_deghost_loss is not None:
            self.deghost_loss = SegmentationLoss(uresnet_deghost, uresnet_deghost_loss)

        # Initialize the segmentation/PPN losses
        if self.segmentation == "uresnet":
            assert not (
                (uresnet_loss is not None) and (uresnet_ppn_loss is not None)
            ), (
                "If the segmentation is using UResNet, can provide either "
                "`uresnet_loss` or `uresnet_ppn_loss` configuration block."
            )
            if uresnet_loss is not None:
                self.uresnet_loss = SegmentationLoss(uresnet, uresnet_loss)
            elif uresnet_ppn_loss is not None:
                self.uresnet_ppn_loss = UResNetPPNLoss(
                    **uresnet_ppn, **uresnet_ppn_loss
                )

        # Initialize the graph-SPICE loss
        if (
            self.fragmentation is not None
            and "graph_spice" in self.fragmentation
            and graph_spice_loss is not None
        ):
            self.graph_spice_loss = GraphSPICELoss(graph_spice, graph_spice_loss)

        # Initialize the GraPA lossses
        self.grappa_losses = {
            "shower": grappa_shower_loss,
            "track": grappa_track_loss,
            "particle": grappa_particle_loss,
            "inter": grappa_inter_loss,
        }
        for stage, config in self.grappa_losses.items():
            if getattr(self, f"{stage}_aggregation") == "grappa" and config is not None:
                name = f"grappa_{stage}_loss"
                setattr(self, name, GrapPALoss(config))

    @property
    def modes(self):
        """Dictionary of (stage, modes) pairs which determine which options
        are available to each of the the reconstruction stage.

        Returns
        -------
        Dict[str, Tuple(str)]
            Dictionary of (stage, modes) pairs to be used
        """
        return dict(self._modes)

    def forward(
        self,
        seg_label=None,
        ppn_label=None,
        clust_label=None,
        clust_label_adapt=None,
        coord_label=None,
        graph_label=None,
        meta=None,
        ghost=None,
        ghost_pred=None,
        segmentation=None,
        seg_pred=None,
        **output,
    ):
        """Run the full chain output through the full chain loss.

        Parameters
        ----------
        seg_label : TensorBatch, optional
            (N, 1 + D + 1) Tensor of segmentation labels
            - 1 is the segmentation label
        ppn_label : TensorBatch, optional
            (N, 1 + D + N_l) Tensor of PPN labels
        clust_label : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels
            - N_c is is the number of cluster labels
        clust_label_adapt : TensorBatch, optional
            (N, 1 + D + N_c) Tensor of cluster labels adapted to seg predictions
            - N_c is is the number of cluster labels
        coord_label : TensorBatch, optional
            (P, 1 + D + 8) Tensor of start/end point labels for each
            true particle in the image
        graph_label : EdgeIndexTensor, optional
            (2, E) Tensor of edges that correspond to physical
            connections between true particle in the image
        meta : Meta, optional
            Image metadata information
        ghost : TensorBatch, optional
            (N, 2) Tensor of logits from the deghosting model
        ghost_pred : TensorBatch, optional
            (N,) Tensor of ghost predictions
        segmentation : TensorBatch, optional
            (N, N_c) Tensor of logits from the segmentation model
        seg_pred : TensorBatch, optional
            (N) Semantic prediction for each point
        **output : dict, optional
            Additional outputs of the reconstruction chain
        """
        # Initialize the loss output
        self.result = {"accuracy": 1.0, "loss": 0.0, "num_losses": 0}

        # Apply the deghosting loss
        if self.deghosting == "uresnet" and hasattr(self, "deghost_loss"):
            # Convert segmentation labels to ghost labels
            ghost_label_tensor = seg_label.tensor.clone()
            ghost_label_tensor[:, SHAPE_COL] = (
                seg_label.tensor[:, SHAPE_COL] == GHOST_SHP
            )
            ghost_label = TensorBatch(ghost_label_tensor, seg_label.counts)

            # Store the loss dictionary
            res_deghost = self.deghost_loss(seg_label=ghost_label, segmentation=ghost)
            self.update_result(res_deghost, "ghost")

            # Restrict the segmentation labels and segmentation outputs
            # to true non-ghosts (do not apply deghosting loss twice)
            if segmentation is not None:
                # Find the index of true non-ghosts in the pred non-ghosts
                deghost_index = ghost_pred.tensor == 0
                seg_label_t = seg_label.tensor[deghost_index]
                index = seg_label_t[:, SHAPE_COL] < GHOST_SHP

                seg_label = TensorBatch(
                    seg_label_t[index],
                    batch_size=seg_label.batch_size,
                    has_batch_col=True,
                )
                segmentation = TensorBatch(segmentation.tensor[index], seg_label.counts)

        # Apply the segmentation and point proposal loss
        if self.segmentation == "uresnet":
            # Adapt the cluster labels to those corresponding to the
            # reconstructed semantic segmentation of the image
            clust_label = clust_label_adapt

            # Store the loss dictionary, if requested
            if hasattr(self, "uresnet_loss"):
                res_seg = self.uresnet_loss(
                    seg_label=seg_label, segmentation=segmentation
                )
                self.update_result(res_seg, "uresnet")

            elif hasattr(self, "uresnet_ppn_loss"):
                res_seg = self.uresnet_ppn_loss(
                    seg_label=seg_label,
                    ppn_label=ppn_label,
                    clust_label=clust_label,
                    segmentation=segmentation,
                    **output,
                )
                self.update_result(res_seg)

        # Apply the Graph-SPICE loss
        if (
            self.fragmentation is not None
            and "graph_spice" in self.fragmentation
            and hasattr(self, "graph_spice_loss")
        ):
            # Prepare Graph-SPICE loss input
            loss_dict = {}
            for key, value in output.items():
                if key.startswith("graph_spice_"):
                    loss_dict[key.replace("graph_spice_", "")] = value

            # Store the loss dictionary
            seg_pred = TensorBatch(seg_pred.tensor[:, None], clust_label.counts)
            res_gs = self.graph_spice_loss(
                seg_label=seg_pred, clust_label=clust_label, **loss_dict
            )
            self.update_result(res_gs, "graph_spice")

        # Apply the aggregation losses
        for stage in self.grappa_losses.keys():
            name = f"grappa_{stage}_loss"
            if getattr(self, f"{stage}_aggregation") == "grappa" and hasattr(
                self, name
            ):
                # Prepare the input to the loss function
                prefix = f"{stage}_fragment" if stage != "inter" else "particle"
                loss_dict = {}
                for k, v in output.items():
                    if f"{prefix}_" in k:
                        loss_dict[k.split(f"{prefix}_")[-1]] = v

                # Store the loss dictionary
                res_grappa = getattr(self, name)(
                    clust_label=clust_label,
                    coord_label=coord_label,
                    graph_label=graph_label,
                    meta=meta,
                    **loss_dict,
                )
                self.update_result(res_grappa, f"grappa_{stage}")

        return self.result

    def update_result(self, result, prefix=None):
        """Update loss and accuracy using the output of one of the module.

        Parameters
        ----------
        result : dict
            Dictionary output of the module
        prefix : str, optional
            Prefix to preface the loss output keys with
        """
        # Update the loss, accuracy and count
        self.result["loss"] += result["loss"]
        self.result["accuracy"] = (
            self.result["accuracy"] * self.result["num_losses"] + result["accuracy"]
        ) / (self.result["num_losses"] + 1)
        self.result["num_losses"] += 1

        # Store the results
        if prefix is None:
            result.pop("loss")
            result.pop("accuracy")
            self.result.update(result)
        else:
            self.result.update({f"{prefix}_{k}": v for k, v in result.items()})


def process_chain_config(self, dump_config=False, **parameters):
    """Process the full chain configuration and dump it.

    Parameters
    ----------
    dump_config : bool, default False
        Whether to dump the chain configuration in the log file or not
    **parameters : dict
        Dictionary of chain configuration parameters
    """
    # Store the modes for each step of the reconstruction. Make sure that
    # that the configuration is recognized.
    for module, valid_modes in self.modes.items():
        valid_modes = (None, "label", *valid_modes)
        assert module in parameters, (
            f"Must configure the {module} stage in the `chain` block. "
            f"The {module} mode should be one of {valid_modes}."
        )
        assert parameters[module] in valid_modes, (
            f"The {module} mode should be one of {valid_modes}. "
            f"Received '{parameters[module]}' instead."
        )
        setattr(self, module, parameters[module])

    # Do some logic checks on the chain parameters
    assert (
        not self.charge_rescaling or self.deghosting
    ), "Charge rescaling cannot be done without deghosting."
    assert self.charge_rescaling == self.deghosting or (
        self.deghosting == "uresnet" and self.charge_rescaling != "label"
    ), (
        "Label-based charge rescaling can only be done in conjunction "
        "with label-based deghosting."
    )
    assert not self.segmentation == "label" or self.deghosting in [None, "label"], (
        "Can only use labels for segmentation if labels are also used for "
        "remove ghost points."
    )
    assert (
        not self.point_proposal == "ppn" or self.segmentation == "uresnet"
    ), "For PPN to work, need the UResNet segmentation backbone."
    assert (
        not self.shower_primary == "gnn" or self.shower_aggregation == "gnn"
    ), "To use GNN for shower primaries, must aggregate showers with it."
    assert not self.particle_aggregation or (
        not self.shower_aggregation and not self.track_aggregation
    ), "Use particle aggregator or shower/track aggregators, not both."

    # Dump the chain configuration, if requested
    if dump_config:
        logger.info(f"Full chain configuration:")
        for k, v in parameters.items():
            v = v if v is not None else "null"
            logger.info(f"  {k:<27}: {v}")
        logger.info("")



"""Post-processor module template.

Use this template as a basis to build your own post-processor. A post-processor
takes the output of the reconstruction and either
- Sets additional reconstruction attributes (e.g. direction estimates)
- Adds entirely new data products (e.g. trigger time)
"""

# Add the imports specific to this module here
# import ...

# Must import the post-processor base class
from spine.post.base import PostBase

# Must list the post-processor(s) here to be found by the factory.
# You must also add it to the list of imported modules in the
# `spine.post.factories`!
__all__ = ["TemplateProcessor"]


class TemplateProcessor(PostBase):
    """Description of what the post-processor is supposed to be doing."""

    name = "template"  # Name used to call the post-processor in the config

    def __init__(self, arg0, arg1, obj_type, run_mode):
        """Initialize the post-processor.

        Parameters
        ----------
        arg0 : type
            Description of arg0
        arg1 : type
            Description of arg1
        obj_type : Union[str, List[str]]
            Types of objects needed in this post-processor (fragments,
            particles and/or interactions). This argument is shared between
            all post-processors. If None, does not load these objects.
        run_mode : str
            One of 'reco', 'truth' or 'both'. Determines what kind of object
            the post-processor has to run on.
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode)

        # Store parameter
        self.arg0 = arg0
        self.arg1 = arg1

        # Add additional required data products
        self.keys["prod"] = True  # Means we must have 'prod' in the dictionary

    def process(self, data):
        """Pass data products corresponding to one entry through the processor.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Fetch the keys you want
        data = data["prod"]

        # Loop over all requested object types
        for key in self.obj_keys:
            # Loop over all objects of that type
            for obj in data[key]:
                # Fetch points attributes
                points = self.get_points(obj)

                # Get another attribute
                sources = obj.sources

                # Do something...

        # Loop over requested specific types of objects
        for key in self.fragment_keys:
            # Do something...
            pass

        for key in self.particle_keys:
            # Do something...
            pass

        for key in self.interaction_keys:
            # Do something...
            pass

        # Return an update or override to the current data product dictionary
        return {}  # Can have no return as well if objects are edited in place



"""Module with a general-purpose geometry class.

This class supports the storage of:
- TPC boundaries
- Optical detector shape/locations
- CRT detector shape/locations

It also provides a plethora of useful functions to query the geometry.
"""

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from scipy.spatial.distance import cdist

from .detector import Box, CRTDetector, OptDetector, TPCDetector


@dataclass
class Geometry(Box):
    """Handles all geometry functions for a collection of box-shaped TPCs with
    a arbitrary set of optical detectors organized in optical volumes and
    auxiliary CRT planes.

    Attributes
    ----------
    name : str
        Name of the detector
    tag : str
        Tag or label for the geometry instance
    version : int
        Version number of the geometry
    tpc : TPCDetector
        TPC detector properties
    optical : OptDetector, optional
        Optical detector properties
    crt : CRTDetector, optional
        CRT detector properties
    gdml : str, optional
        GDML file name associated with the geometry
    crs_files : List[str], optional
        CRS (Charge Readout System) geometry file(s) reference (flow only)
    lrs_file : str, optional
        LRS (Light Readout System) geometry file reference (flow only)
    """

    name: str
    tag: str
    version: str
    tpc: TPCDetector
    optical: Optional[OptDetector] = None
    crt: Optional[CRTDetector] = None
    gdml: Optional[str] = None
    crs_files: Optional[List[str]] = None
    lrs_file: Optional[str] = None

    def __init__(
        self,
        name: str,
        tag: str,
        version: str,
        tpc: Dict[str, Any],
        optical: Optional[Dict[str, Any]] = None,
        crt: Optional[Dict[str, Any]] = None,
        gdml: Optional[str] = None,
        crs_files: Optional[List[str]] = None,
        lrs_file: Optional[str] = None,
    ):
        """Initialize the detector geometry.

        Parameters
        ----------
        name : str
            Name of the detector
        tag : str
            Tag or label for the geometry instance
        version : int
            Version number of the geometry
        tpc : dict
            Detector boundary configuration
        optical : dict, optional
            Optical detector configuration
        crt : dict, optional
            CRT detector configuration
        gdml : str, optional
            GDML file name associated with the geometry
        crs_files : str or list of str, optional
            CRS (Cosmic Ray System) geometry file(s)
        lrs_file : str, optional
            LRS (Light Readout System) geometry file

        Returns
        -------
        np.ndarray
            Lower boundaries of the overall geometry
        np.ndarray
            Upper boundaries of the overall geometry
        """
        # Store basic geometry information
        self.name = name
        self.tag = tag
        self.version = version
        self.gdml = gdml
        self.crs_files = crs_files
        self.lrs_file = lrs_file

        # Load the charge detector boundaries
        self.tpc = TPCDetector(**tpc)
        lower = self.tpc.lower
        upper = self.tpc.upper

        # Load the optical detectors
        if optical is not None:
            self.optical = self.parse_optical(**optical)
            lower = np.minimum(lower, self.optical.lower)
            upper = np.maximum(upper, self.optical.upper)

        # Load the CRT detectors
        if crt is not None:
            self.crt = CRTDetector(**crt)
            lower = np.minimum(lower, self.crt.lower)
            upper = np.maximum(upper, self.crt.upper)

        # Initialize the parent Box
        super().__init__(lower, upper)

    def parse_optical(self, volume: str, **optical: Any) -> OptDetector:
        """Parse the optical detector configuration.

        Parameters
        ----------
        volume : str
            Optical volume boundaries (one of 'tpc' or 'module')
        **optical : dict
            Reset of the optical detector configuration

        Returns
        -------
        OptDetector
            Optical detector object
        """
        # Get the number of optical volumes based on the the volume type
        assert volume in [
            "module",
            "tpc",
        ], "Optical detector positions must be provided by TPC or module."

        if volume == "module":
            offsets = [module.center for module in self.tpc.modules]
        else:
            offsets = [chamber.center for chamber in self.tpc.chambers]

        # Initialize the optical detector object
        return OptDetector(volume, offsets, **optical)

    @dataclass
    class ContDefinition:
        """Helper class to store containment volume definitions.

        Attributes
        ----------
        volumes : List[np.ndarray]
            (N, 3, 2) set of volume boundaries
        use_source : bool
            Whether to use the source of the points to determine the volume
        limit_normals : List[np.ndarray]
            List of limit plane normals
        limit_thresholds : List[float]
            List of limit thresholds along the plane normal
        """

        volumes: np.ndarray
        use_source: bool
        limit_normals: List[np.ndarray]
        limit_thresholds: List[float]

    def get_boundaries(
        self, with_optical: bool = True, with_crt: bool = True
    ) -> np.ndarray:
        """Fetch the overall geometry boundaries, optionally including
        optical and CRT detectors.

        Parameters
        ----------
        with_optical : bool, default True
            Whether to include optical detector boundaries
        with_crt : bool, default True
            Whether to include CRT detector boundaries

        Returns
        -------
        np.ndarray
            Lower and upper boundaries of the overall geometry
        """
        # Get the TPC boundaries
        lower = self.tpc.lower
        upper = self.tpc.upper

        # Expand to include optical and CRT detectors, if requested
        if with_optical:
            assert (
                self.optical is not None
            ), "This geometry does not have optical detectors to include."
            lower = np.minimum(lower, self.optical.lower)
            upper = np.maximum(upper, self.optical.upper)

        if with_crt:
            assert (
                self.crt is not None
            ), "This geometry does not have CRT detectors to include."
            lower = np.minimum(lower, self.crt.lower)
            upper = np.maximum(upper, self.crt.upper)

        return np.vstack((lower, upper)).T

    def get_sources(self, sources: np.ndarray) -> np.ndarray:
        """Converts logical TPC indexes to physical TPC indexes.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of logical [module ID, tpc ID] pairs, one per point

        Returns
        ----------
        np.ndarray
            (N, 2) Array of physical [module ID, tpc ID] pairs, one per point
        """
        # If logical and physical TPCs are aligned, nothing to do
        if self.tpc.det_ids is None:
            return sources.astype(int)

        # Otherwise, map logical to physical
        sources = np.copy(sources)
        sources[:, 1] = self.tpc.det_ids[sources[:, 1].astype(int)]

        return sources

    def get_chambers(self, sources: np.ndarray) -> np.ndarray:
        """Converts logical TPC indexes to unique chamber indexes.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of logical [module ID, tpc ID] pairs, one per point

        Returns
        ----------
        np.ndarray
            (N) Array of physical chamber indexes, one per point
        """
        # Convert the physical indexes to unique chamber indexes
        sources = self.get_sources(sources)
        return sources[:, 0] * self.tpc.num_chambers_per_module + sources[:, 1]

    def get_contributors(self, sources: np.ndarray) -> List[np.ndarray]:
        """Gets the list of [module ID, tpc ID] pairs that contributed to an
        object, as defined in this geometry.

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of [module ID, tpc ID] pairs, one per point

        Returns
        -------
        List[np.ndarray]
            (2, N_t) Pair of arrays: the first contains the list of
            contributing modules, the second of contributing tpcs.
        """
        # Fetch the list of unique logical [module ID, tpc ID] pairs
        sources = np.unique(sources, axis=0)

        # If the logical TPCs differ from the physical TPCs, convert
        if self.tpc.det_ids is not None:
            sources = self.get_sources(sources)
            sources = np.unique(sources, axis=0)

        # Return as a list of physical [module ID, tpc ID] pairs
        return list(sources.T)

    def get_volume_index(
        self, sources: np.ndarray, module_id: int, tpc_id: Optional[int] = None
    ) -> np.ndarray:
        """Gets the list of indices of points that belong to a certain
        detector volume (module or individual TPC).

        Parameters
        ----------
        sources : np.ndarray
            (N, 2) Array of [module ID, tpc ID] pairs, one per point
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (N) Index of points that belong to the requested detector volume
        """
        # If the logical TPCs differ from the physical TPCs, convert
        sources = self.get_sources(sources)

        # Compute and return the index
        if tpc_id is None:
            return np.where(sources[:, 0] == module_id)[0]
        else:
            return np.where((sources == [module_id, tpc_id]).all(axis=-1))[0]

    def get_closest_tpc(self, points: np.ndarray) -> np.ndarray:
        """For each point, find the ID of the closest TPC.

        There is a natural assumption that all TPCs are boxes of identical
        sizes, so that the relative proximitity of a point to a TPC is
        equivalent to its proximity to the TPC center.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        np.ndarray
            (N) List of TPC indexes, one per input point
        """
        # Get the TPC centers
        centers = np.asarray([chamber.center for chamber in self.tpc.chambers])

        # Compute the pair-wise distances between points and TPC centers
        dists = cdist(points, centers)

        # Return the closest center index as the closest centers
        return np.argmin(dists, axis=1)

    def get_closest_module(self, points: np.ndarray) -> np.ndarray:
        """For each point, find the ID of the closest module.

        There is a natural assumption that all modules are boxes of identical
        sizes, so that the relative proximitity of a point to a module is
        equivalent to its proximity to the module center.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        np.ndarray
            (N) List of module indexes, one per input point
        """
        # Get the module centers
        centers = np.asarray([module.center for module in self.tpc.modules])

        # Compute the pair-wise distances between points and module centers
        dists = cdist(points, centers)

        # Return the closest center index as the closest centers
        return np.argmin(dists, axis=1)

    def get_closest_tpc_indexes(self, points: np.ndarray) -> List[np.ndarray]:
        """For each TPC, get the list of points that live closer to it than any
        other TPC in the detector.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        List[np.ndarray]
            List of index of points that belong to each TPC
        """
        # Start by finding the closest TPC to each of the points
        closest_ids = self.get_closest_tpc(points)

        # For each TPC, append the list of point indices associated with it
        tpc_indexes = []
        for t in range(self.tpc.num_chambers):
            tpc_indexes.append(np.where(closest_ids == t)[0])

        return tpc_indexes

    def get_closest_module_indexes(self, points: np.ndarray) -> List[np.ndarray]:
        """For each module, get the list of points that live closer to it
        than any other module in the detector.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates

        Returns
        -------
        List[np.ndarray]
            List of index of points that belong to each module
        """
        # Start by finding the closest TPC to each of the points
        closest_ids = self.get_closest_module(points)

        # For each module, append the list of point indices associated with it
        module_indexes = []
        for m in range(self.tpc.num_modules):
            module_indexes.append(np.where(closest_ids == m)[0])

        return module_indexes

    def get_volume_offsets(
        self, points: np.ndarray, module_id: int, tpc_id: Optional[int] = None
    ) -> np.ndarray:
        """Compute how far each point is from a certain detector volume.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) : Point coordinates
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (N, 3) Offsets w.r.t. to the volume boundaries
        """
        # Compute the axis-wise distances of each point to each boundary
        module = self.tpc[module_id]
        if tpc_id is None:
            boundaries = module.boundaries
        else:
            boundaries = module[tpc_id].boundaries
        dists = points[..., None] - boundaries

        # If a point is between two boundaries, the distance is 0. If it is
        # outside, the distance is that of the closest boundary
        signs = (np.sign(dists[..., 0]) + np.sign(dists[..., 1])) / 2
        offsets = signs * np.min(np.abs(dists), axis=-1)

        return offsets

    def get_min_volume_offset(
        self, points: np.ndarray, module_id: int, tpc_id: Optional[int] = None
    ) -> np.ndarray:
        """Get the minimum offset to apply to a point cloud to bring it
        within the boundaries of a volume.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) : Point coordinates
        module_id : int
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module. If not specified, the volume
            offsets are estimated w.r.t. the module itself

        Returns
        -------
        np.ndarray
            (3) Offsets w.r.t. to the volume location
        """
        # Compute the distance for each point, get the maximum necessary offset
        offsets = self.get_volume_offsets(points, module_id, tpc_id)
        offsets = offsets[np.argmax(np.abs(offsets), axis=0), np.arange(3)]

        return offsets

    def translate(
        self,
        points: np.ndarray,
        source_id: int,
        target_id: int,
        factor: Optional[Union[float, np.ndarray]] = None,
    ) -> np.ndarray:
        """Moves a point cloud from one module to another one

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates
        source_id: int
            Module ID from which to move the point cloud
        target_id : int
            Module ID to which to move the point cloud
        factor : Union[float, np.ndarray], optional
            Multiplicative factor to apply to the offset. This is necessary if
            the points are not expressed in detector coordinates

        Returns
        -------
        np.ndarray
            (N, 3) Set of translated point coordinates
        """
        # If the source and target are the same, nothing to do here
        if target_id == source_id:
            return np.copy(points)

        # Fetch the inter-module shift
        offset = self.tpc[target_id].center - self.tpc[source_id].center
        if factor is not None:
            offset *= factor

        # Translate
        return points + offset

    def split(
        self,
        points: np.ndarray,
        target_id: int,
        sources: Optional[np.ndarray] = None,
        meta: Optional[Any] = None,
    ) -> Tuple[np.ndarray, List[np.ndarray]]:
        """Migrate all points to a target module, organize them by module ID.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Set of point coordinates
        target_id : int
            Module ID to which to move the point cloud
        sources : np.ndarray, optional
            (N, 2) Array of [module ID, tpc ID] pairs, one per voxel
        meta : Meta, optional
            Meta information about the voxelized image. If provided, the
            points are assumed to be provided in voxel coordinates.

        Returns
        -------
        np.ndarray
            (N, 3) Shifted set of points
        List[np.ndarray]
            List of index of points that belong to each module
        """
        # Check that the target ID exists
        assert (
            target_id > -1 and target_id < self.tpc.num_modules
        ), "Target ID should be in [0, N_modules[."

        # Get the module ID of each of the input points
        convert = False
        if sources is not None:
            # If sources are provided, simply use that
            module_indexes = []
            for m in range(self.tpc.num_modules):
                module_indexes.append(np.where(sources[:, 0] == m)[0])

        else:
            # If the points are expressed in pixel coordinates, translate
            if meta is not None:
                convert = True
                points = meta.to_cm(points, center=True)

            # If not provided, find module each point belongs to by proximity
            module_indexes = self.get_closest_module_indexes(points)

        # Now shifts all points that are not in the target
        for module_id, module_index in enumerate(module_indexes):
            # If this is the target module, nothing to do here
            if module_id == target_id:
                continue

            # Shift the coordinates
            points[module_index] = self.translate(
                points[module_index], module_id, target_id
            )

        # Bring the coordinates back to pixels, if they were shifted
        if meta is not None and convert:
            points = meta.to_px(points, floor=True)

        return points, module_indexes

    def check_containment(
        self,
        definition: ContDefinition,
        points: np.ndarray,
        sources: Optional[np.ndarray] = None,
        allow_multi_module: bool = False,
        summarize: bool = True,
    ):
        """Check whether a point cloud comes within some distance of the boundaries
        of a certain subset of detector volumes, depending on the mode.

        Must define containment volumes first using `define_containment_volumes`.

        Parameters
        ----------
        definition : ContDefinition
            Containment volume definition to check against
        points : np.ndarray
            (N, 3) Set of point coordinates
        sources : np.ndarray, optional
            (S, 2) : List of [module ID, tpc ID] pairs that created the
            point cloud
        allow_multi_module : bool, default `False`
            Whether to allow points to span multiple modules
        summarize : bool, default `True`
            If `True`, only returns a single flag for the whole cloud.
            Otherwise, returns a boolean array corresponding to each point.

        Returns
        -------
        Union[bool, np.ndarray]
            `True` if the points are contained, `False` if not
        """
        # If sources are provided, only consider source volumes
        if definition.use_source:
            # Get the contributing TPCs
            assert sources is not None and len(points) == len(
                sources
            ), "Need to provide sources to make a source-based check."
            contributors = self.get_contributors(sources)
            if not allow_multi_module and len(np.unique(contributors[0])) > 1:
                return False

            # Define the smallest box containing all contributing TPCs
            # TODO: this is not ideal, should check each source separately?
            index = contributors[0] * self.tpc.num_chambers_per_module + contributors[1]
            volume = self.merge_volumes(definition.volumes[index])
            volumes = [volume]

        else:
            volumes = definition.volumes

        # Loop over volumes, make sure the cloud is contained in at least one
        if summarize:
            contained = False
            for v in volumes:
                if (points > v[:, 0]).all() and (points < v[:, 1]).all():
                    contained = True
                    break
        else:
            contained = np.zeros(len(points), dtype=bool)
            for v in volumes:
                contained |= (points > v[:, 0]).all(axis=1) & (points < v[:, 1]).all(
                    axis=1
                )

        # If requested, check points against the active volume limits
        if not summarize or contained:
            for normal, threshold in zip(
                definition.limit_normals, definition.limit_thresholds
            ):
                active = np.dot(points, normal) < threshold
                if summarize:
                    active = active.all()
                contained &= active

        return contained

    def define_containment_volumes(
        self,
        margin: Union[float, List[float], np.ndarray],
        cathode_margin: Optional[float] = None,
        mode: str = "module",
        include_limits: bool = True,
    ) -> ContDefinition:
        """This function defines a list of volumes to check containment against.

        If the containment is checked against a constant volume, it is a lot
        more efficient to call this function once and call `check_containment`
        reapitedly after.

        Parameters
        ----------
        margin : Union[float, List[float], np.array]
            Minimum distance from a detector wall to be considered contained:
            - If float: distance buffer is shared between all 6 walls
            - If [x,y,z]: distance is shared between pairs of walls facing
              each other and perpendicular to a shared axis
            - If [[x_low,x_up], [y_low,y_up], [z_low,z_up]]: distance is
              specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        mode : str, default 'module'
            Containement criterion (one of 'global', 'module', 'tpc'):
            - If 'tpc', makes sure it is contained within a single TPC
            - If 'module', makes sure it is contained within a single module
            - If 'detector', makes sure it is contained within in the detector
            - If 'source', use the origin of voxels to determine which TPC(s)
              contributed to them, and define volumes accordingly
        include_limits : bool, default True
            If `True`, the TPC active region limits are checked against

        Returns
        -------
        ContDefinition
            Object containing the list of containment volumes, and other
            information to check containment against
        """
        # Translate the margin parameter to a (3,2) matrix
        margin = np.array(margin)
        if len(margin.shape) == 0:
            margin = np.full((3, 2), margin)
        elif len(margin.shape) == 1:
            assert len(margin) == 3, "Must provide one value per axis."
            margin = np.tile(margin, (2, 1)).T
        else:
            assert np.array(margin).shape == (3, 2), "Must provide two values per axis."

        # Establish the volumes to check against
        cont_volumes = []
        cont_use_source = False
        if mode in ["tpc", "source"]:
            for m, module in enumerate(self.tpc):
                for t, tpc in enumerate(module):
                    vol = self.adapt_volume(
                        tpc.boundaries, margin, cathode_margin, m, t
                    )
                    cont_volumes.append(vol)
            cont_use_source = mode == "source"

        elif mode == "module":
            for module in self.tpc:
                vol = self.adapt_volume(module.boundaries, margin)
                cont_volumes.append(vol)

        elif mode == "detector":
            vol = self.adapt_volume(self.tpc.boundaries, margin)
            cont_volumes.append(vol)

        else:
            raise ValueError(f"Containement check mode not recognized: {mode}.")

        cont_volumes = np.array(cont_volumes)

        # Establish active region limits to check against, if requested
        limit_normals, limit_thresholds = [], []
        if include_limits and self.tpc.limits is not None:
            assert np.all(margin == margin[0, 0]), (
                "No clear way to include active region limit checks when "
                "margins are different in different axes. Abort."
            )
            limit_margin = margin[0, 0]
            for limit in self.tpc.limits:
                limit_normals.append(limit.norm)
                limit_thresholds.append(limit.boundary - limit_margin)

        # Return containment conditions as a ContDefinition object
        return self.ContDefinition(
            volumes=cont_volumes,
            use_source=cont_use_source,
            limit_normals=limit_normals,
            limit_thresholds=limit_thresholds,
        )

    def adapt_volume(
        self,
        ref_volume: np.ndarray,
        margin: np.ndarray,
        cathode_margin: Optional[float] = None,
        module_id: Optional[int] = None,
        tpc_id: Optional[int] = None,
    ):
        """Apply margins from a given volume. Takes care of subtleties
        associated with the cathode, if needed.

        Parameters
        ----------
        ref_volume : np.ndarray
            (3, 2) Array of volume boundaries
        margin : np.ndarray
            Minimum distance from a detector wall to be considered contained as
            [[x_low,x_up], [y_low,y_up], [z_low,z_up]], i.e. distance is
            specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        module_id : int, optional
            ID of the module
        tpc_id : int, optional
            ID of the TPC within the module

        Returns
        -------
        np.ndarray
            (3, 2) Updated array of volume boundaries
        """
        # Reduce the volume according to the margin
        volume = np.copy(ref_volume)
        volume[:, 0] += margin[:, 0]
        volume[:, 1] -= margin[:, 1]

        # If a cathode margin is provided, adapt the cathode wall differently
        if cathode_margin is not None:
            assert (
                module_id is not None and tpc_id is not None
            ), "Module and TPC ID must be provided when using a cathode margin."
            axis = self.tpc[module_id][tpc_id].drift_axis
            side = self.tpc[module_id][tpc_id].cathode_side

            flip = (-1) ** side
            volume[axis, side] += flip * (cathode_margin - margin[axis, side])

        return volume

    @staticmethod
    def merge_volumes(volumes: np.ndarray) -> np.ndarray:
        """Given a list of volumes and their boundaries, find the smallest box
        that encompass all volumes combined.

        Parameters
        ----------
        volumes : np.ndarray
            (N, 3, 2) List of volume boundaries

        Returns
        -------
        np.ndarray
            (3, 2) Boundaries of the combined volume
        """
        volume = np.empty((3, 2))
        volume[:, 0] = np.min(volumes, axis=0)[:, 0]
        volume[:, 1] = np.max(volumes, axis=0)[:, 1]

        return volume



"""Docstring inheritance utilities."""


def inherit_docstring(*parents):
    """Inherits docstring attributes of a parent class.

    Only handles numpy-style docstrings.

    Parameters
    ----------
    *parents : List[object]
        Parent class(es) to inherit attributes from

    Returns
    -------
    callable
        Class with updated docstring
    """

    def inherit(obj):
        tab = "    "
        underline = "----"
        header = f"Attributes\n{tab}----------\n"

        # If there is no attribute or method yet, add the header
        if not header in obj.__doc__:
            obj.__doc__ += f"\n\n{tab}{header}"

        # Get the parent attribute docstring block
        prestr = ""
        for parent in parents:
            docstr = parent.__doc__
            substr = docstr.split(header)[-1].rstrip() + "\n"
            if len(substr.split(underline)) > 1:
                substr = substr.split(underline)[0].split("\n")[:-1]
                substr = "".join(substr).rstrip()

            prestr += substr

        # Append it to the relevant block
        split_doc = obj.__doc__.split(header)
        obj.__doc__ = split_doc[0] + header + prestr + split_doc[1]

        return obj

    return inherit



"""PyTorch runtime utilities for tensor operations and memory management.

This module provides conditional PyTorch utilities that gracefully handle
PyTorch unavailability with sensible fallbacks or clear error messages.
"""

from ..conditional import TORCH_AVAILABLE, torch

__all__ = [
    "manual_seed",
    "cuda_is_available",
    "cuda_mem_info",
    "cuda_max_memory_allocated",
    "is_tensor",
    "distributed_barrier",
    "require_torch",
]


def manual_seed(seed):
    """Set torch manual seed if torch is available."""
    if TORCH_AVAILABLE:
        torch.manual_seed(seed)


def cuda_is_available():
    """Check if CUDA is available."""
    return TORCH_AVAILABLE and torch.cuda.is_available()


def cuda_mem_info():
    """Get CUDA memory info if available."""
    if TORCH_AVAILABLE:
        return torch.cuda.mem_get_info()
    return (0, 0)  # Return (used, total) as 0,0 if not available


def cuda_max_memory_allocated():
    """Get max CUDA memory allocated if available."""
    if TORCH_AVAILABLE and torch.cuda.is_available():
        return torch.cuda.max_memory_allocated()
    return 0


def is_tensor(obj):
    """Check if object is a torch tensor."""
    return TORCH_AVAILABLE and torch.is_tensor(obj)


def distributed_barrier():
    """Call distributed barrier if available."""
    if TORCH_AVAILABLE and torch.distributed.is_available():
        torch.distributed.barrier()


def require_torch(operation="this operation"):
    """Raise informative error when torch is required but not available."""
    if not TORCH_AVAILABLE:
        raise ImportError(
            f"PyTorch is required for {operation}. "
            "Install with: pip install spine-ml[model]"
        )



"""Functions which initialize activation and normalization layers
used generically by torch models.
"""

from copy import deepcopy

from torch import nn

from spine.utils.factory import instantiate

__all__ = ["act_factory", "norm_factory"]


def act_dict():
    """Dictionary of recognized activation functions."""

    activations = {
        "none": nn.Identity,
        "elu": nn.ELU,
        "relu": nn.ReLU,
        "lrelu": nn.LeakyReLU,
        "prelu": nn.PReLU,
        "selu": nn.SELU,
        "celu": nn.CELU,
        "tanh": nn.Tanh,
        "sigmoid": nn.Sigmoid,
        "softmax": nn.Softmax,
        "softplus": nn.Softplus,
        "mish": nn.Mish,
    }

    return activations


def norm_dict():
    """Dictionary of recognized normalization functions."""

    normalizations = {
        "none": nn.Identity,
        "batch_norm": nn.BatchNorm1d,
        "instance_norm": nn.InstanceNorm1d,
        "group_norm": nn.GroupNorm,
        "layer_norm": nn.LayerNorm,
    }

    return normalizations


def act_factory(cfg):
    """Instantiates an activation layer.

    Parameters
    ----------
    cfg : dict
        Activation layer configuration

    Return
    ------
    object
        Instantiated activation layer
    """
    return instantiate(act_dict(), cfg)


def norm_factory(cfg, num_features=None):
    """Instantiates a normalization layer.

    Parameters
    ----------
    cfg : dict
        Normalization layer configuration
    num_features : int
        Number of features to normalize

    Return
    ------
    object
        Instantiated normalization layer
    """
    return instantiate(norm_dict(), cfg, num_features=num_features)



"""Module to build GNN message passing submodules."""

from spine.utils.factory import instantiate, module_dict

from . import layer

__all__ = ["node_layer_factory", "edge_layer_factory", "global_layer_factory"]


def node_layer_factory(cfg, node_in, edge_in, glob_in):
    """Instantiates a GNN node update layer from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        GNN node update layer configuration
    node_in : int
        Number of input node features
    edge_in : int
        Number of input edge features
    glob_in : int
        Number of input global graph features

    Returns
    -------
    object
        Instantiated GNN node update layer
    """
    layer_dict = module_dict(layer, pattern="Node")
    return instantiate(
        layer_dict, cfg, node_in=node_in, edge_in=edge_in, glob_in=glob_in
    )


def edge_layer_factory(cfg, node_in, edge_in, glob_in):
    """Instantiates a GNN edge update layer from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        GNN edge update layer configuration
    node_in : int
        Number of input node features
    edge_in : int
        Number of input edge features
    glob_in : int
        Number of input global graph features

    Returns
    -------
    object
        Instantiated GNN edge update layer
    """
    layer_dict = module_dict(layer, pattern="Edge")
    return instantiate(
        layer_dict, cfg, node_in=node_in, edge_in=edge_in, glob_in=glob_in
    )


def global_layer_factory(cfg, node_in, glob_in):
    """Instantiates a GNN global update layer from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        GNN global update layer configuration
    node_in : int
        Number of input node features
    glob_in : int
        Number of input global graph features

    Returns
    -------
    object
        Instantiated GNN global update layer
    """
    layer_dict = module_dict(layer, pattern="Global")
    return instantiate(layer_dict, cfg, node_in=node_in, edge_in=edge_in)



"""Module with all the backbone components of UResNet.

Contains the following components:
  - `UResNetEncoder`: Encoder component of UResNet
  - `UResNetDecoder`: Decoder component of UResNet
  - `UResNet`: Full encoder/decoder architecture of UResNet
"""

from typing import List

import MinkowskiEngine as ME
import torch

from .act_norm import act_factory, norm_factory
from .blocks import ASPP, CascadeDilationBlock, ResNetBlock
from .configuration import setup_cnn_configuration

__all__ = ["UResNetEncoder", "UResNetDecoder", "UResNet"]


class UResNetEncoder(torch.nn.Module):
    """Vanilla UResNet encoder.

    See :func:`setup_cnn_configuration` for available parameters.
    """

    def __init__(self, cfg):
        """Initialize the encoder.

        Parameters
        ----------
        cfg : dict
            Encoder configuration block
        """
        # Initialize the parent class
        super().__init__()

        # Process the configuration
        setup_cnn_configuration(self, **cfg)

        # Initialize the input layer
        self.input_layer = ME.MinkowskiConvolution(
            in_channels=self.num_input,
            out_channels=self.num_filters,
            kernel_size=self.input_kernel,
            stride=1,
            dimension=self.dim,
            bias=self.allow_bias,
        )

        # Initialize encoder
        self.encoding_conv = []
        self.encoding_block = []
        for i, F in enumerate(self.num_planes):
            m = []
            for _ in range(self.reps):
                m.append(
                    ResNetBlock(
                        F,
                        F,
                        dimension=self.dim,
                        activation=self.act_cfg,
                        normalization=self.norm_cfg,
                        bias=self.allow_bias,
                    )
                )
            m = torch.nn.Sequential(*m)
            self.encoding_block.append(m)
            m = []
            if i < self.depth - 1:
                m.append(norm_factory(self.norm_cfg, F))
                m.append(act_factory(self.act_cfg))
                m.append(
                    ME.MinkowskiConvolution(
                        in_channels=self.num_planes[i],
                        out_channels=self.num_planes[i + 1],
                        kernel_size=2,
                        stride=2,
                        dimension=self.dim,
                        bias=self.allow_bias,
                    )
                )
            m = torch.nn.Sequential(*m)
            self.encoding_conv.append(m)
        self.encoding_conv = torch.nn.Sequential(*self.encoding_conv)
        self.encoding_block = torch.nn.Sequential(*self.encoding_block)

    def forward(self, x):
        """Pass a tensor through the encoder.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        encoder_tensors : List[ME.SparseTensor]
            List of intermediate tensors (taken between encoding block and
            convolution) from the encoder half
        final_tensor : ME.SparseTensor
            Feature tensor at deepest layer
        """
        x = self.input_layer(x)
        encoder_tensors = [x]
        for i, layer in enumerate(self.encoding_block):
            x = self.encoding_block[i](x)
            encoder_tensors.append(x)
            x = self.encoding_conv[i](x)

        result = {"encoder_tensors": encoder_tensors, "final_tensor": x}

        return result


class UResNetDecoder(torch.nn.Module):
    """Vanilla UResNet Decoder.

    See :func:`setup_cnn_configuration` for available parameters.
    """

    def __init__(self, cfg):
        """Initialize the decoder.

        Parameters
        ----------
        cfg : dict
            Decoder configuration block
        """
        # Initialize the parent class
        super().__init__()

        # Process the configuration
        setup_cnn_configuration(self, **cfg)

        # Initialize decoder
        self.decoding_block = []
        self.decoding_conv = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(norm_factory(self.norm_cfg, self.num_planes[i + 1]))
            m.append(act_factory(self.act_cfg))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.num_planes[i + 1],
                    out_channels=self.num_planes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.dim,
                    bias=self.allow_bias,
                )
            )
            m = torch.nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ResNetBlock(
                        self.num_planes[i] * (2 if j == 0 else 1),
                        self.num_planes[i],
                        dimension=self.dim,
                        activation=self.act_cfg,
                        normalization=self.norm_cfg,
                        bias=self.allow_bias,
                    )
                )
            m = torch.nn.Sequential(*m)
            self.decoding_block.append(m)
        self.decoding_block = torch.nn.Sequential(*self.decoding_block)
        self.decoding_conv = torch.nn.Sequential(*self.decoding_conv)

    def forward(self, final, encoder_tensors):
        """Pass a tensor through the decoder.

        Parameters
        ----------
        final : ME.SparseTensor
            Output of the encoder
        encoder_tensors : List[ME.SparseTensor]
            List of tensors from each depth of the encoder

        Returns
        -------
        List[ME.SparseTensor]
            List of feature tensors in decoding path at each spatial resolution
        """
        decoder_tensors = []
        x = final
        for i, layer in enumerate(self.decoding_conv):
            encoder_tensor = encoder_tensors[-i - 2]
            x = layer(x)
            x = ME.cat(encoder_tensor, x)
            x = self.decoding_block[i](x)
            decoder_tensors.append(x)

        return decoder_tensors


class UResNet(torch.nn.Module):
    """Vanilla UResNet with access to intermediate feature planes.

    See :func:`setup_cnn_configuration` for available parameters.
    """

    def __init__(self, cfg):
        """Initialize the UResNet backbone.

        Parameters
        ----------
        cfg : dict
            Decoder configuration block
        """
        # Initialize the parent class
        super().__init__()

        # Process the configuration
        setup_cnn_configuration(self, **cfg)

        # Initialize the encoder/decoder blocks of the UResNet model
        self.encoder = UResNetEncoder(cfg)
        self.decoder = UResNetDecoder(cfg)

    def forward(self, input_data):
        """Pass a tensor through the UResNet backbone.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        encoder_tensors : List[ME.SparseTensor]
            List of intermediate tensors (taken between encoding block and
            convolution) from the encoder half
        decoder_tensors : List[ME.SparseTensor]
            List of feature tensors in decoding path at each spatial resolution
        final_tensor : ME.SparseTensor
            Feature tensor at deepest layer
        """
        # Cast the input data to a sparse tensor
        coords = input_data[:, 0 : self.dim + 1].int()
        features = input_data[:, self.dim + 1 :]
        x = ME.SparseTensor(features, coordinates=coords)

        # Pass it through the encoder
        encoder_output = self.encoder(x)
        encoder_tensors = encoder_output["encoder_tensors"]
        final_tensor = encoder_output["final_tensor"]

        # Pass it through the decoder
        decoder_tensors = self.decoder(final_tensor, encoder_tensors)

        # Return
        res = {
            "encoder_tensors": encoder_tensors,
            "decoder_tensors": decoder_tensors,
            "final_tensor": final_tensor,
        }

        return res



"""True particle children counting module."""

from collections import Counter, defaultdict

import numpy as np

from spine.post.base import PostBase
from spine.utils.globals import GHOST_SHP, PID_LABELS

__all__ = ["ChildrenProcessor"]


class ChildrenProcessor(PostBase):
    """Count the number of children of a given particle, using the particle
    hierarchy information from :class:`ParticleGraphParser`.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "children_count"

    # Alternative allowed names of the post-processor
    aliases = ("count_children",)

    def __init__(self, mode="shape", obj_type="particle"):
        """Initialize the children counting parameters.

        Parameters
        ----------
        mode : str, default 'shape'
            Attribute name to categorize children. This will count each child
            object for different category separately.
        """
        # Initialize the parent class
        super().__init__(obj_type, "truth")

        # Store the counting mode
        self.mode = mode
        if self.mode == "shape":
            self.num_classes = GHOST_SHP
        elif self.mode == "pid":
            self.num_classes = len(PID_LABELS) - 1
        else:
            raise ValueError(
                f"Child counting mode not recognized: {mode}. Must be "
                "one of 'shape' or 'pid'."
            )

    def process(self, data):
        """Count children of each true particle in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the requested data products
        for k in self.obj_keys:
            # Build parent-child relationships using simple dictionaries
            nodes = {}  # node_id -> attribute
            children = defaultdict(list)  # parent_id -> [child_ids]

            # First pass: collect all nodes and their attributes
            for obj in data[k]:
                nodes[obj.orig_id] = getattr(obj, self.mode)

            # Second pass: build parent-child relationships
            for obj in data[k]:
                parent_id = obj.parent_id
                if parent_id in nodes and int(parent_id) != int(
                    obj.orig_id
                ):  # Avoid self-loops
                    children[parent_id].append(obj.orig_id)

            # Count children for each object
            for obj in data[k]:
                child_ids = children[obj.orig_id]
                counter = Counter()
                counter.update([nodes[child_id] for child_id in child_ids])
                children_counts = np.zeros(self.num_classes, dtype=np.int64)
                for key, val in counter.items():
                    children_counts[key] = val

                obj.children_counts = children_counts



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import fps, knn
from torch_scatter import scatter_mean

from .lovasz import StableBCELoss, lovasz_hinge, lovasz_softmax_flat

# Collection of Miscellaneous Loss Functions not yet implemented in Pytorch.


def unique_label_torch(label):
    _, label2, cts = torch.unique(label, return_inverse=True, return_counts=True)
    return label2, cts


def iou_batch(pred: torch.BoolTensor, labels: torch.BoolTensor, eps=0.0):
    """
    pred: N x C
    labels: N x C (one-hot)
    """
    intersection = (pred & labels).float().sum(0)
    union = (pred | labels).float().sum(0)
    iou = (intersection + eps) / (union + eps)  # We smooth our devision to avoid 0/0

    return iou.mean()


class LovaszHingeLoss(torch.nn.modules.loss._Loss):

    def __init__(self, reduction="none"):
        super(LovaszHingeLoss, self).__init__(reduction=reduction)

    def forward(self, logits, targets):
        num_clusters = targets.shape[1]
        return lovasz_hinge(
            logits.T.view(num_clusters, 1, -1), targets.T.view(num_clusters, 1, -1)
        )


class LovaszSoftmaxWithLogitsLoss(torch.nn.modules.loss._Loss):

    def __init__(self, reduction="none"):
        super(LovaszSoftmaxWithLogitsLoss, self).__init__(reduction=reduction)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, logits, targets):
        probs = self.softmax(logits)
        return lovasz_softmax_flat(probs, targets)


def find_cluster_means(features, labels):
    """
    For a given image, compute the centroids mu_c for each
    cluster label in the embedding space.
    Inputs:
        features (torch.Tensor): the pixel embeddings, shape=(N, d) where
        N is the number of pixels and d is the embedding space dimension.
        labels (torch.Tensor): ground-truth group labels, shape=(N, )
    Returns:
        cluster_means (torch.Tensor): (n_c, d) tensor where n_c is the number of
        distinct instances. Each row is a (1,d) vector corresponding to
        the coordinates of the i-th centroid.
    """
    centroids = scatter_mean(features, labels, dim=0)
    return centroids


def intra_cluster_loss(features, cluster_means, labels, margin=1.0):
    """
    Computes the intra-cluster loss between an embedding point cloud and
    a set of attractor points.

    <labels> must range between 0 to the number of <cluster_means>, otherwise
    the loss will be underestimated as <scatter_mean> zero value placeholders.
    """
    from torch_scatter import scatter_mean

    x = features[:, None, :]
    mu = cluster_means[None, :, :]
    l = torch.clamp(torch.norm(x - mu, dim=-1) - margin, min=0) ** 2
    l = torch.gather(l, 1, labels.view(-1, 1)).squeeze()

    if len(l.size()) and len(labels.size()):
        intra_loss = torch.mean(scatter_mean(l, labels))
        return intra_loss
    else:
        # print('intra_cluster_loss', l.size(), labels.size())
        return 0.0


def inter_cluster_loss(cluster_means, margin=0.2):
    inter_loss = 0.0
    n_clusters = len(cluster_means)
    if n_clusters < 2:
        # Inter-cluster loss is zero if there only one instance exists for
        # a semantic label.
        return 0.0
    else:
        indices = torch.triu_indices(cluster_means.shape[0], cluster_means.shape[0], 1)
        dist = squared_distances(cluster_means, cluster_means)
        return torch.pow(
            torch.clamp(2.0 * margin - dist[indices[0, :], indices[1, :]], min=0), 2
        ).mean()


def regularization_loss(cluster_means):
    return torch.mean(torch.norm(cluster_means, dim=1))


def margin_smoothing_loss(sigma, sigma_means, labels, margin=0):
    from torch_scatter import scatter_mean

    x = sigma[:, None]
    mu = sigma_means[None, :]
    l = torch.sqrt(torch.clamp(torch.abs(x - mu) - margin, min=0) ** 2 + 1e-6)
    l = torch.gather(l, 1, labels.view(-1, 1)).view(-1)
    loss = torch.mean(scatter_mean(l, labels))
    return loss


def get_probs(embeddings, margins, labels, eps=1e-6):
    from torch_scatter import scatter_mean

    device = embeddings.device
    n = labels.shape[0]
    centroids = find_cluster_means(embeddings, labels)
    sigma = scatter_mean(margins.squeeze(), labels)
    num_clusters = labels.unique().shape[0]

    # Compute spatial term
    em = embeddings[:, None, :]
    centroids = centroids[None, :, :]
    sqdists = ((em - centroids) ** 2).sum(-1)

    p = sqdists / (2.0 * sigma.view(1, -1) ** 2)
    p = torch.clamp(torch.exp(-p), min=eps, max=1 - eps)
    logits = torch.logit(p, eps=eps)
    eye = torch.eye(len(labels.unique()), dtype=torch.float32, device=device)
    targets = eye[labels]
    loss_tensor = nn.BCEWithLogitsLoss(reduction="none")(logits, targets)
    loss = loss_tensor.mean(dim=0).mean()
    with torch.no_grad():
        acc = iou_batch(logits > 0, targets.bool())
    smoothing_loss = margin_smoothing_loss(
        margins.squeeze(), sigma.detach(), labels, margin=0
    )
    p = torch.gather(p, 1, labels.view(-1, 1))
    return loss, smoothing_loss, p.squeeze(), acc


def multivariate_kernel(centroid, log_sigma, Lprime, eps=1e-8):
    def f(x):
        N = x.shape[0]
        L = torch.zeros(3, 3)
        tril_indices = torch.tril_indices(row=3, col=3, offset=-1)
        L[tril_indices[0], tril_indices[1]] = Lprime
        sigma = torch.exp(log_sigma) + eps
        L += torch.diag(sigma)
        cov = torch.matmul(L, L.T)
        dist = torch.matmul((x - centroid), torch.inverse(cov))
        dist = torch.bmm(dist.view(N, 1, -1), (x - centroid).view(N, -1, 1)).squeeze()
        probs = torch.exp(-dist)
        return probs

    return f


def bhattacharyya_distance_matrix(v1, v2, eps=1e-8):
    x1, s1 = v1[:, :3], v1[:, 3].view(-1)
    x2, s2 = v2[:, :3], v1[:, 3].view(-1)
    g1 = torch.ger(s1**2, 1.0 / (s2**2 + eps))
    g2 = g1.t()
    dist = squared_distances(x1.contiguous(), x2.contiguous())
    denom = 1.0 / (eps + s1.unsqueeze(1) ** 2 + s2**2)
    out = 0.25 * torch.log(0.25 * (g1 + g2 + 2)) + 0.25 * dist / denom
    return out


def squared_distances(v1, v2):
    v1_2 = v1.unsqueeze(1).expand(v1.size(0), v2.size(0), v1.size(1)).double()
    v2_2 = v2.unsqueeze(0).expand(v1.size(0), v2.size(0), v1.size(1)).double()
    return torch.pow(v2_2 - v1_2, 2).sum(2)


def bhattacharyya_coeff_matrix(v1, v2, eps=1e-6):
    x1, s1 = v1[:, :3], v1[:, 3].view(-1)
    x2, s2 = v2[:, :3], v1[:, 3].view(-1)
    g1 = torch.ger(s1**2, 1.0 / (s2**2 + eps))
    g2 = g1.t()
    dist = torch.cidst(x1.contiguous(), x2.contiguous())
    denom = 1.0 / (eps + s1.unsqueeze(1) ** 2 + s2**2)
    out = 0.25 * torch.log(0.25 * (g1 + g2 + 2)) + 0.25 * dist / denom
    out = torch.exp(-out)
    return out


def get_graphspice_logits(
    sp_emb,
    ft_emb,
    cov,
    groups,
    sp_centroids,
    ft_centroids,
    eps=0.001,
    compute_accuracy=True,
):

    device = sp_emb.device
    cov_means = find_cluster_means(cov, groups)

    # Compute spatial term
    sp_emb_tmp = sp_emb[:, None, :]
    sp_centroids_tmp = sp_centroids[None, :, :]
    sp_cov = torch.clamp(cov_means[:, 0][None, :], min=eps)
    sp_sqdists = ((sp_emb_tmp - sp_centroids_tmp) ** 2).sum(-1) / (sp_cov**2)

    # Compute feature term
    ft_emb_tmp = ft_emb[:, None, :]
    ft_centroids_tmp = ft_centroids[None, :, :]
    ft_cov = torch.clamp(cov_means[:, 0][None, :], min=eps)
    ft_sqdists = ((ft_emb_tmp - ft_centroids_tmp) ** 2).sum(-1) / (ft_cov**2)

    # Compute joint kernel score
    pvec = torch.exp(-sp_sqdists - ft_sqdists)
    # probs = (1-pvec).index_put((torch.arange(groups.shape[0]), groups),
    #     torch.gather(pvec, 1, groups.view(-1, 1)).squeeze())
    logits = torch.logit(pvec)

    acc = None
    eye = torch.eye(len(groups.unique()), dtype=torch.float32, device=device)
    targets = eye[groups]
    if compute_accuracy:
        acc = iou_batch(logits > 0, targets.bool())

    return logits, acc, targets



"""MCS angular resolution calibration module.

This module uses high energy muon tracks to evaluate the base angular resolution
of the tracks as a basis to correct the scattering angle spread in the MCS-based
KE estimation used in SPINE.
"""

from functools import partial

import numpy as np

from spine.ana.base import AnaBase
from spine.utils.globals import MUON_PID
from spine.utils.mcs import ANGLE_METHODS, mcs_angles, mcs_angles_proj
from spine.utils.tracking import get_track_segments

__all__ = ["MCSCalibAna"]


class MCSCalibAna(AnaBase):
    """This analysis script estimates the angular resolution of high energy
    muon chunks to calibrate the MCS-based KE reconstruction.

    This script proceeds through the following steps:
    - Identify high energy muons
    - Chunk them up in various sizes
    - Estimate the 3D polar angle between successive chunks
    """

    # Name of the analysis script (as specified in the configuration)
    name = "calib_mcs"

    # Axes names used to store direction information
    _axes = ("x", "y", "z")

    # Projection names used to store projected angle information
    _projs = ("yz", "xz", "xy")

    def __init__(
        self,
        min_ke,
        segment_length,
        segment_method="bin_pca",
        anchor_point=True,
        angle_method="atan2",
        time_window=None,
        run_mode="truth",
        truth_point_mode="points",
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        min_ke : float
            Kinetic energy (in MeV) above which a muon is included in the estimate
        segment_length : Union[float, List[float]]
            Segment length in the units that specify the coordinates. If array,
            will produce angular measurements for each segmentation length
        segment_method : str, default 'step_next'
            Method used to segment the track (one of 'step', 'step_next'
            or 'bin_pca')
        anchor_point : bool, default True
            Weather or not to collapse end point onto the closest track point
        angle_method : str, default 'atan2'
            Angular reconstruction method
        time_window : List[float], optional
            Time within which the particle must occur to be used for calibration.
            This is only used with true instances to remove out-of-time objects
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(
            "particle", run_mode, truth_point_mode=truth_point_mode, **kwargs
        )

        # Store the kinematic cut
        self.min_ke = min_ke

        # Store the segmentation parameters
        self.segment_length = segment_length
        if np.isscalar(segment_length):
            self.segment_length = [segment_length]

        # Partially initialize the segmentation algorithm
        self.get_track_segments = partial(
            get_track_segments, method=segment_method, anchor_point=anchor_point
        )

        # Store the angle computation parameters
        assert angle_method in ANGLE_METHODS, (
            f"Angular reconstruction method not recognized: {angle_method}. "
            f"Must be one of {ANGLE_METHODS.keys()}"
        )
        self.angle_method = ANGLE_METHODS[angle_method]

        # Store the time window, if provided
        assert (
            time_window is None or run_mode == "truth"
        ), "Cannot enforce time window containment on reconstructed objects."
        assert time_window is None or (
            not np.isscalar(time_window) and len(time_window) == 2
        ), "If a time window is provided, it must be a list of two scalars."
        self.time_window = time_window

        # Initialize the CSV writer(s) you want (one per segment length)
        for prefix in self.prefixes:
            for sl in self.segment_length:
                self.initialize_writer(f"{prefix}_mcs_sl{sl}")

    def process(self, data):
        """Pass data products corresponding to one entry through the analysis.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle keys
        for prefix in self.prefixes:
            # Loop over particles
            for part in data[f"{prefix}_particles"]:
                # Check PID and kinetic energy
                if part.pid != MUON_PID or part.ke < self.min_ke:
                    continue

                # Enfore time, if requested
                if self.time_window is not None:
                    if part.t < self.time_window[0] or part.t > self.time_window[1]:
                        continue

                # Loop over segment lengths
                for sl in self.segment_length:
                    # Segment the track
                    segments, dirs, lengths = self.get_track_segments(
                        self.get_points(part), sl, part.start_point
                    )
                    if len(segments) < 1:
                        continue

                    # Compute angles and directions
                    angles = mcs_angles(dirs, self.angle_method)
                    angles_proj = mcs_angles_proj(dirs, self.angle_method)
                    dirs = (dirs[:-1] + dirs[1:]) / 2

                    # Record the size of the smallest of the two segments
                    counts = np.array([len(seg) for seg in segments])
                    min_counts = np.min(np.vstack((counts[:-1], counts[1:])), axis=0)

                    # Record the distance between two successive segment centers
                    distances = (lengths[:-1] + lengths[1:]) / 2

                    # Write to the file
                    file_name = f"{prefix}_mcs_sl{sl}"
                    base = {"ke": part.ke}
                    if part.is_truth:
                        base["time"] = part.t
                    for i, angle in enumerate(angles):
                        dir_dict = {
                            f"dir_{a}": dirs[i, j] for j, a in enumerate(self._axes)
                        }
                        proj_dict = {
                            f"angle_{p}": angles_proj[i, j]
                            for j, p in enumerate(self._projs)
                        }
                        self.append(
                            file_name,
                            **base,
                            **dir_dict,
                            **proj_dict,
                            angle=angle,
                            min_count=min_counts[i],
                            distance=distances[i],
                        )



"""JIT compilation decorators using Numba for performance optimization."""

import inspect
from functools import wraps

import numba as nb
import numpy as np

from .conditional import TORCH_AVAILABLE, torch


def numbafy(cast_args=[], list_args=[], keep_torch=False, ref_arg=None):
    """Function which wraps a `numba` function with some checks on the input
    to make the relevant conversions to numpy where necessary.

    Parameters
    ----------
    cast_args : list(str), optional
        List of arguments to be cast to numpy
    list_args : list(str), optional
        List of arguments which need to be cast to a numba typed list
    keep_torch : bool, default False
        Make the output a torch object, if the reference argument is one
    ref_arg : str, optional
        Reference argument used to assign a type and device to the torch output

    Returns
    -------
    callable
        Wrapped function which ensures input type compatibility with numba
    """

    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            # Convert the positional arguments in args into
            # key:value pairs in kwargs
            keys = list(inspect.signature(fn).parameters.keys())
            for i, val in enumerate(args):
                kwargs[keys[i]] = val

            # Extract the default values for the remaining parameters
            for key, val in inspect.signature(fn).parameters.items():
                if key not in kwargs and val.default != inspect.Parameter.empty:
                    kwargs[key] = val.default

            # If a torch output is request, register the input dtype and
            # device location
            if keep_torch:
                assert (
                    ref_arg in kwargs
                ), "Must provide the `ref_arg` to cast back to torch."
                dtype, device = None, None
                if TORCH_AVAILABLE and isinstance(kwargs[ref_arg], torch.Tensor):
                    dtype = kwargs[ref_arg].dtype
                    device = kwargs[ref_arg].device

            # If the cast data is not a numpy array, cast it
            for arg in cast_args:
                assert arg in kwargs, (
                    f"Argument `{arg}` appears in `cast_args` but does "
                    "not appear in the function arguments."
                )
                if not isinstance(kwargs[arg], np.ndarray):
                    if TORCH_AVAILABLE and isinstance(kwargs[arg], torch.Tensor):
                        kwargs[arg] = kwargs[arg].detach().cpu().numpy()
                    else:
                        raise TypeError(
                            f"Can only convert torch.Tensor to numpy when PyTorch is available. "
                            f"Got a request to cast {type(kwargs[arg])} instead."
                        )

            # If there is a reflected list in the input, type it
            for arg in list_args:
                assert arg in kwargs, (
                    f"Argument `{arg}` appears in `list_args` but does "
                    "not appear in the function arguments."
                )
                kwargs[arg] = nb.typed.List(kwargs[arg])

            # Get the output
            ret = fn(**kwargs)
            if keep_torch and dtype and TORCH_AVAILABLE:
                if isinstance(ret, np.ndarray):
                    ret = torch.tensor(ret, dtype=dtype, device=device)
                elif isinstance(ret, list):
                    ret = [torch.tensor(r, dtype=dtype, device=device) for r in ret]
                else:
                    raise TypeError("Return type not recognized, cannot cast to torch.")
            return ret

        return inner

    return outer



"""Defines class used to assign orphaned points to a sensible cluster."""

import numpy as np

from spine.math.cluster import DBSCAN
from spine.math.neighbors import KNeighborsClassifier, RadiusNeighborsClassifier

__all__ = ["OrphanAssigner"]


class OrphanAssigner:
    """Clustering orphan assignment.

    This class takes care of finding the best match cluster ID for points that
    have not found a suitable group in the upstream clustering.

    This is a wrapper class for two classes:
    - :class:`KNeighborsClassifier`
    - :class:`RadiusNeighborsClassifier`
    """

    def __init__(self, mode, assign_all=True, **kwargs):
        """Initialize the orphan assigner.

        Parameters
        ----------
        mode : str
            Orphan assignment mode, one of 'knn' or 'radius'
        assign_all : bool, default True
            If `True`, force assign all orphans to a cluster. In the 'knn' mode,
            this is guaranteed, provided there is at least one labeled point.
            In the 'radius' mode, this uses DBSCAN for outliers.
        **kwargs : dict
            Arguments to pass to the underlying classifier function
        """
        # Initialize the classifier
        self.mode = mode
        if mode == "knn":
            self.classifier = KNeighborsClassifier(**kwargs)
        elif mode == "radius":
            self.classifier = RadiusNeighborsClassifier(**kwargs)
        else:
            raise ValueError(
                "The orphan assignment mode must be one of 'knn' or "
                f"'radius'. Got '{mode}' instead."
            )

        # Store the extra parameter
        self.assign_all = assign_all

        # If needed, initialize DBSCAN
        if mode == "radius" and assign_all:
            radius = kwargs.get("radius")
            metric = kwargs.get("metric", "euclidean")
            self.dbscan = DBSCAN(
                eps=radius, min_samples=1, metric=metric, p=self.classifier.p
            )

    def __call__(self, X, y):
        """Place-holder for a function which assigns labels to orphan points.

        Parameters
        ----------
        X : np.ndarray
            (N, 3) Coordinates of the points in the image
        y : np.ndarray
            (N) Labels of the points (-1 if orphaned)

        Returns
        -------
        np.ndarray
            (M) Labels assigned to each of the orphans
        """
        # Create a mask to identify labeled and orphaned points
        orphan_mask = y == -1
        orphan_index = np.where(orphan_mask)[0]
        valid_index = np.where(~orphan_mask)[0]

        # Assign orphan points using the neighbor classifier
        labels, orphan_update = self.classifier.fit_predict(
            X[valid_index], y[valid_index], X[orphan_index]
        )

        y_updated = y.copy()
        y_updated[orphan_index] = labels
        orphan_index = orphan_index[orphan_update]

        # If required, assign stragglers using DBSCAN
        if len(orphan_index) and self.mode == "radius" and self.assign_all:
            # Get the assignment for each of the orphaned points
            update = self.dbscan.fit_predict(X[orphan_index])

            # Update the labels accordingly
            offset = np.max(y_updated) + 1
            y_updated[orphan_index] = offset + update

        return y_updated



import numpy as np
import torch
import torch.nn as nn


def digamma_evd_loss(alpha, y):
    """
    Bayes risk loss for Dirichlet prior Evidential Learning

    INPUTS:
        - alpha (FloatTensor): N x C concentration parameters,
        where C is the number of class labels.
        - y (FloatTensor): N x C one-hot encoded class labels

    RETURNS:
        - loss (FloatTensor): N x 1 non-reduced loss for each example.
    """
    S = alpha.sum(dim=1, keepdim=True)
    loss = torch.sum((torch.digamma(S) - torch.digamma(alpha)) * y, dim=1)
    return loss


def sumsq_evd_loss(alpha, y):
    """
    Negative log loss for Dirichlet prior evidential learning.

    INPUTS:
        - alpha (FloatTensor): N x C concentration parameters,
        where C is the number of class labels.
        - y (FloatTensor): N x C one-hot encoded class labels

    RETURNS:
        - loss (FloatTensor): N x 1 non-reduced loss for each example.
    """
    S = alpha.sum(dim=1, keepdim=True)
    prediction_err = (y - alpha / S) ** 2
    variance = alpha * (S - alpha) / (S * S * (S + 1.0))
    loss = torch.sum(prediction_err + variance, dim=1)
    return loss


def nll_evd_loss(alpha, y):
    """
    Negative log loss for Dirichlet prior evidential learning.

    INPUTS:
        - alpha (FloatTensor): N x C concentration parameters,
        where C is the number of class labels.
        - y (FloatTensor): N x C one-hot encoded class labels

    RETURNS:
        - loss (FloatTensor): N x 1 non-reduced loss for each example.
    """
    S = alpha.sum(dim=1, keepdim=True)
    loss = torch.sum(y * (torch.log(S) - torch.log(alpha)), dim=1)
    return loss


def evd_kl_divergence(alpha, beta=None):
    """
    KL Divergence between Dir(p|alpha) and Dir(p|beta), where
    alpha and beta are Dirichlet concentration parameters.

    INPUTS:
        - alpha (FloatTensor): N x C concentration parameters
        - beta (FloatTensor): N x C concentration parameters. In case of
        truth labels, this is a one-hot encoded class label tensor.
        If None, this will compute the KL Divergence between Dir(p|alpha)
        and Dir(p|1), which is the uniform distribution over C classes.

    RETURNS:
        - loss (FloatTensor): N x 1 non-reduced kl divergence loss.
    """
    device = alpha.device
    S_alpha = torch.sum(alpha, dim=1)
    if beta is None:
        beta = torch.ones([1, alpha.shape[1]], device=device)
    S_beta = torch.sum(beta, dim=1)
    loss = torch.lgamma(S_alpha) - torch.lgamma(S_beta)
    loss -= torch.sum(torch.lgamma(alpha), dim=1)
    A = (alpha - beta) * (torch.digamma(alpha) - torch.digamma(S_alpha.view(-1, 1)))
    loss += torch.sum(A, dim=1)
    return loss


def evd_loss_dict():
    loss_fn = {
        "edl_digamma": digamma_evd_loss,
        "edl_sumsq": sumsq_evd_loss,
        "edl_nll": nll_evd_loss,
    }
    return loss_fn


def evd_loss_factory(name):
    losses = evd_loss_dict()
    if name not in losses:
        raise Exception("Unknown evd loss algorithm name provided: %s" % name)
    return losses[name]


class EVDLoss(nn.Module):
    """
    Base class for loss used in the paper:
    Sensoy et. al., Evidential Deep Learning to Quantify
    Classification Uncertainty
    """

    def __init__(
        self,
        evd_loss_name,
        reduction="none",
        T=50000,
        one_hot=True,
        num_classes=5,
        mode="concentration",
    ):
        super(EVDLoss, self).__init__()
        self.T = T  # Total epoch counts for which to anneal kld component.
        self.evd_loss = evd_loss_factory(evd_loss_name)
        self.kld_loss = evd_kl_divergence
        self.reduction = reduction
        self.one_hot = one_hot
        self.num_classes = num_classes
        self.mode = mode

    def forward(self, alpha, labels, t=0):

        device = alpha.device
        if self.one_hot:
            eye = torch.eye(self.num_classes).to(device=device)
            y = eye[labels.long()]
        else:
            y = labels

        if self.mode != "concentration":
            evidence = alpha
            alpha = evidence + 1.0

        annealing = min(1.0, float(t) / self.T)

        evd_loss = self.evd_loss(alpha, y)
        alpha_tilde = y + (1 - y) * alpha
        kld_loss = self.kld_loss(alpha_tilde)

        loss = evd_loss + annealing * kld_loss
        if self.reduction == "none":
            return loss
        elif self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        else:
            raise Exception("Unknown reduction method %s provided" % self.reduction)

    def __str__(self):
        str_format = (
            "EVDLoss(name={}, reduction={}, one_hot={}, num_classes={}, mode={})"
        )
        return str_format.format(
            self.evd_loss_name,
            self.reduction,
            self.one_hot,
            self.num_classes,
            self.mode,
        )


def nll_regression_loss(logits, targets, eps=1e-6):
    """
    Negative log loss for Dirichlet prior evidential learning.

    INPUTS:
        - alpha (FloatTensor): N x C concentration parameters,
        where C is the number of class labels.
        - y (FloatTensor): N x 1 regression targets

    RETURNS:
        - loss (FloatTensor): N x 1 non-reduced loss for each example.
    """
    gamma, nu, alpha, beta = logits[:, 0], logits[:, 1], logits[:, 2], logits[:, 3]
    omega = 2.0 * beta * (1.0 + nu)
    nll = (
        0.5 * (np.log(np.pi) - torch.log(nu + 1e-5))
        - alpha * torch.log(omega)
        + (alpha + 0.5) * torch.log(nu * (targets - gamma) ** 2 + omega)
        + torch.lgamma(alpha)
        - torch.lgamma(alpha + 0.5)
    )
    return torch.clamp(nll, min=0)


def kld_regression_loss(logits, targets, eps=1e-6):
    gamma, nu, alpha = logits[:, 0], logits[:, 1], logits[:, 2]
    loss = torch.abs(targets - gamma + eps) * (2.0 * nu + alpha)
    return loss


def kld_evd_l2_loss(logits, targets, eps=1e-6):
    gamma, nu, alpha = logits[:, 0], logits[:, 1], logits[:, 2]
    loss = torch.pow(targets - gamma + eps, 2) * (2.0 * nu + alpha)
    return loss


def kl_nig(logits, targets, eps=0.01):
    gamma, nu, alpha = logits[:, 0], logits[:, 1], logits[:, 2]

    error = torch.abs(targets - gamma + 1e-6)

    kl = (
        0.5 * (1.0 + eps + 0.001) / (nu + 0.001)
        - 0.5
        - torch.lgamma(alpha / (1.0 + eps))
        + (alpha - (1.0 + eps)) * torch.digamma(alpha)
    )

    loss = kl * error
    return loss


class EDLRegressionLoss(nn.Module):

    def __init__(
        self,
        reduction="none",
        w=0.0,
        kl_mode="evd",
        one_hot="True",
        mode="concentration",
        eps=1e-6,
        T=50000,
        logspace=False,
    ):
        super(EDLRegressionLoss, self).__init__()
        self.reduction = reduction
        self.one_hot = one_hot
        self.mode = mode
        self.eps = eps
        self.nll_loss = nll_regression_loss
        self.kl_mode = kl_mode
        if self.kl_mode == "evd":
            self.kld_loss = kld_regression_loss
        elif self.kl_mode == "kl":
            self.kld_loss = kl_nig
        elif self.kl_mode == "evd_l2":
            self.kld_loss = kld_evd_l2_loss
        else:
            raise ValueError("Unrecognized KL Divergence Error Loss")
        self.w = w
        self.T = T
        self.logspace = logspace

    def forward(self, logits, targets, iteration=None):

        logits = logits.view(-1, 4)
        if self.logspace:
            p_true = torch.log(targets + 1e-6)
        else:
            p_true = targets

        assert len(targets.shape) == 1

        if iteration is not None:
            annealing = min(1.0, float(iteration) / self.T)
        else:
            annealing = self.w

        nll_loss = self.nll_loss(logits, p_true, eps=self.eps)
        kld_loss = self.kld_loss(logits, p_true, eps=self.eps)

        return nll_loss + annealing * kld_loss, nll_loss



import numba as nb
import numpy as np
from scipy.interpolate import UnivariateSpline

import spine.math as sm


def get_track_length(
    coordinates: nb.float32[:, :],
    segment_length: nb.float32 = None,
    point: nb.float32[:] = None,
    method: str = "bin_pca",
    anchor_point: bool = True,
    min_count: int = 10,
    spline_smooth: float = None,
) -> nb.float32:
    """Given a set of point coordinates associated with a track and one of its
    end points, compute its length.

    Parameters
    ----------
    coordinates : np.ndarray
        (N, 3) Coordinates of the points that make up the track
    segment_length : float, optional
        Segment length in the units that specify the coordinates
    point : np.ndarray, optional
        (3) An end point of the track
    method : str, default 'bin_pca'
        Method used to compute the track length (one of 'displacement', 'step',
        'step_next', 'bin_pca' or 'spline')
    anchor_point : bool, default True
        Weather or not to collapse end point onto the closest track point
    min_count : int, default 10
        Minimum number of points in a segment to use it to evaluate the
        direction of the next step along the track.
    spline_smooth : float, optional
        The smoothing factor to be used in spline regression, when used

    Returns
    -------
    float
       Total length of the track
    """
    if method == "displacement":
        # Project points along the principal component, compute displacement
        track_dir = sm.decomposition.principal_components(coordinates)[0]
        pcoordinates = np.dot(coordinates, track_dir)

        return np.max(pcoordinates) - np.min(pcoordinates)

    if method in ["step", "step_next", "bin_pca"]:
        # Segment the track and sum the segment lengths
        seg_lengths = get_track_segments(
            coordinates, segment_length, point, method, anchor_point, min_count
        )[-1]

        return np.sum(seg_lengths)

    elif method == "splines":
        # Fit point along the track with a spline, compute spline length
        return get_track_spline(coordinates, segment_length, spline_smooth)[-1]

    else:
        raise ValueError(f"Track length estimation method not recognized: `{method}`.")


@nb.njit(cache=True)
def check_track_orientation(
    coordinates: nb.float32[:, :],
    values: nb.float32[:],
    start_point: nb.float32[:],
    end_point: nb.float32[:],
    method: str = "local",
    anchor_points: bool = True,
    local_radius: nb.float32 = 5,
    segment_method: str = "step_next",
    segment_length: nb.float32 = 5,
    segment_min_count: int = 10,
) -> bool:
    """Given a set of track point coordinates and the track end points, checks
    which end point is most likely to be the correct start, based on the
    rate of energy deposition in the track.

    Parameters
    ----------
    coordinates : np.ndarray
        (N, 3) Coordinates of the points that make up the track
    values : np.ndarray
        (N) Values associated with each point
    start_point : np.ndarray
        (3) Start point of the track
    end_point : np.ndarray
        (3) End point of the track
    method : str, default 'local'
        Method used to orient the track (one of 'local', 'gradient')
    local_radius : float, default 5
        Radius around the end points to used to evaluate the local dE/dx
    anchor_points : bool, default True
        Weather or not to collapse end point onto the closest track point
    segment_method : str, default 'step_next'
        Method used to segment the track when using the 'gradient' method
    segment_length : float, default 5
        Segment length when using the 'gradient' method
    segment_min_count : int, default 10
        Minimum number of points in a segment when using the 'gradient' method

    Returns
    -------
    bool
       Returns `True` if the start point provided is correct, `False`
       if the end point is more likely to be the start point.
    """
    if method == "local":
        # If requested, anchor the end points to the closest track points
        end_points = np.vstack((start_point, end_point))
        if anchor_points:
            dist_mat = sm.distance.cdist(end_points, coordinates)
            end_ids = sm.argmin(dist_mat, axis=1)
            end_points = coordinates[end_ids]

        # Compute the local dE/dx around each end, pick the end with the lowest
        start_dedx = np.sum(values[dist_mat[0] < local_radius]) / local_radius
        end_dedx = np.sum(values[dist_mat[1] < local_radius]) / local_radius

        return start_dedx < end_dedx

    elif method == "gradient":
        # Compute the track gradient with respect to either ends
        grad_start = get_track_deposition_gradient(
            coordinates,
            values,
            start_point,
            segment_length,
            segment_method,
            anchor_points,
            segment_min_count,
        )[0]
        grad_end = get_track_deposition_gradient(
            coordinates,
            values,
            end_point,
            segment_length,
            segment_method,
            anchor_points,
            segment_min_count,
        )[0]

        # Compute the deposition gradient as an average of the two
        gradient = (grad_start - grad_end) / 2.0

        return bool(gradient >= 0.0)

    else:
        raise ValueError("Track orientation method not recognized")


@nb.njit(cache=True)
def get_track_deposition_gradient(
    coordinates: nb.float32[:, :],
    values: nb.float32[:],
    start_point: nb.float32[:],
    segment_length: nb.float32 = 5.0,
    method: str = "step_next",
    anchor_point: bool = True,
    min_count: int = 10,
) -> (nb.float32, nb.float32[:], nb.float32[:], nb.float32[:]):
    """Given a set of point coordinates and their values associated with a
    track and a start point, compute the deposition gradient with respect to
    the start point.

    Parameters
    ----------
    coordinates : np.ndarray
        (N, 3) Coordinates of the points that make up the track
    values : np.ndarray
        (N) Values associated with each point
    start_point : np.ndarray
        (3) End point of the track
    segment_length : float, default 5
        Segment length in the units that specify the coordinates
    method : str, default 'step_next'
        Method used to segment the track (one of 'step', 'step_next'
        or 'bin_pca')
    anchor_point : bool, default True
        Weather or not to collapse end point onto the closest track point
    min_count : int, default 10
        Minimum number of points in a segment for it to be valid. If not valid,
        the dedx value for that segment is not used to compute the gradient.

    Returns
    -------
    gradient : float
       Deposition gradient along the track from the start point
    segment_dedxs : np.ndarray
       (S) Array of energy/charge deposition rate values
    segment_rrs : np.ndarray
       (S) Array of residual ranges (center of the segment w.r.t. end point)
    segment_lengths : np.ndarray
       (S) Array of segment lengths
    """
    # Compute the track segment dedxs
    seg_dedxs, _, seg_rrs, _, _, seg_lengths = get_track_segment_dedxs(
        coordinates,
        values,
        start_point,
        segment_length,
        method,
        anchor_point,
        min_count,
    )

    valid_index = np.where(seg_dedxs > -1)[0]
    if not len(valid_index):
        return 0.0, seg_dedxs, seg_rrs, seg_lengths

    seg_dedxs = seg_dedxs[valid_index]
    seg_rrs = seg_rrs[valid_index]

    # Compute the dE/dx gradient = Cov(x,y) / Var(x)
    cov = np.cov(seg_rrs, seg_dedxs)
    gradient = cov[0, 1] / cov[0, 0] if cov[0, 0] > 0.0 else 0.0

    return gradient, seg_dedxs, seg_rrs, seg_lengths


@nb.njit(cache=True)
def get_track_segment_dedxs(
    coordinates: nb.float32[:, :],
    values: nb.float32[:],
    end_point: nb.float32[:] = None,
    segment_length: nb.float32 = 5.0,
    method: str = "step_next",
    anchor_point: bool = True,
    min_count: int = 10,
) -> (
    nb.float32[:],
    nb.float32[:],
    nb.float32[:],
    nb.float32[:],
    nb.float32[:],
    nb.float32[:],
):
    """Given a set of point coordinates and their values associated with a
    track and one of its end points, compute the energy/charge deposition rate
    as a function of the residual range.

    Parameters
    ----------
    coordinates : np.ndarray
        (N, 3) Coordinates of the points that make up the track
    values : np.ndarray
        (N) Values associated with each point
    end_point : np.ndarray, optional
        (3) End point of the track
    segment_length : float, default 5.
        Segment length in the units that specify the coordinates
    method : str, default 'step_next'
        Method used to segment the track (one of 'step', 'step_next'
        or 'bin_pca')
    anchor_point : bool, default True
        Weather or not to collapse end point onto the closest track point
    min_count : int, default 10
        Minimum number of points in a segment for it to be valid. If not valid,
        the dedx value returned for the segment is -1.

    Returns
    -------
    seg_dedxs : np.ndarray
       (S) Array of energy/charge deposition rate values
    seg_errs : np.ndarray
       (S) Array of uncertainties on the energy/charge deposition rate
    seg_rrs : np.ndarray
       (S) Array of residual ranges (center of the segment w.r.t. end point)
    seg_clusts : List[np.ndarray]
       (S) List of indexes which correspond to each segment cluster of points
    seg_dirs : np.ndarray
       (S, 3) Array of segment direction vectors
    seg_lengths : np.ndarray
       (S) Array of segment lengths
    """
    # Get the segment indexes and their lengths
    seg_clusts, seg_dirs, seg_lengths = get_track_segments(
        coordinates, segment_length, end_point, method, anchor_point, min_count
    )

    # Compute the dQdxs and residual ranges
    seg_dedxs = np.empty(len(seg_clusts), dtype=np.float32)
    seg_errs = np.empty(len(seg_clusts), dtype=np.float32)
    seg_rrs = np.empty(len(seg_clusts), dtype=np.float32)
    residual_range = 0.0
    for i, seg in enumerate(seg_clusts):
        # Compute the rate of energy/charge deposition
        # If the segment has insufficient content, return dummy values
        dx = seg_lengths[i]
        if len(seg) >= min_count and dx > 0.0:
            de = np.sum(values[seg])
            dde = np.std(values[seg]) * np.sqrt(len(seg))
            seg_dedxs[i] = de / dx
            seg_errs[i] = dde / dx
        else:
            seg_dedxs[i] = -1.0
            seg_errs[i] = -1.0

        # Compute the residual_range
        seg_rrs[i] = residual_range + dx / 2.0
        residual_range += dx

    return seg_dedxs, seg_errs, seg_rrs, seg_clusts, seg_dirs, seg_lengths


@nb.njit(cache=True)
def get_track_segments(
    coordinates: nb.float32[:, :],
    segment_length: nb.float32,
    point: nb.float32[:] = None,
    method: str = "step_next",
    anchor_point: bool = True,
    min_count: int = 10,
) -> (nb.types.List(nb.int64[:]), nb.float32[:, :], nb.float64[:]):
    """Given a set of point coordinates associated with a track and one of its
    end points, divide the track into segments of the requested length.

    Parameters
    ----------
    coordinates : np.ndarray
        (N, 3) Coordinates of the points that make up the track
    segment_length : float
        Segment length in the units that specify the coordinates
    point : np.ndarray, optional
        (3) A preferred end point of the track from which to start
    method : str, default 'step_next'
        Method used to segment the track (one of 'step', 'step_next'
        or 'bin_pca')
    anchor_point : bool, default True
        Weather or not to collapse end point onto the closest track point
    min_count : int, default 10
        Minimum number of points in a segment to use it to evaluate the
        direction of the next step along the track.

    Returns
    -------
    segment_clusts : List[np.ndarray]
       (S) List of indexes which correspond to each segment cluster of points
    segment_dirs : np.ndarray
       (S, 3) Array of segment direction vectors
    segment_lengths : np.ndarray
       (S) Array of segment lengths
    """
    if method == "step" or method == "step_next":
        # Determine which point to start stepping from
        if point is not None:
            start_point = point
            if anchor_point:
                start_id = np.argmin(
                    sm.distance.cdist(np.atleast_2d(point), coordinates)
                )
                start_point = coordinates[start_id]
        else:
            # If not specified, pick a random end point of the track
            start_id = sm.distance.farthest_pair(coordinates)[0]
            start_point = coordinates[start_id]

        # Step through the track iteratively
        empty_list = nb.typed.List.empty_list
        seg_start = np.copy(start_point)
        seg_clusts = empty_list(np.empty(0, dtype=np.int64))
        seg_dirs_l = empty_list(np.empty(0, dtype=coordinates.dtype))
        seg_lengths_l = empty_list(nb.float64)
        left_index = np.arange(len(coordinates))
        while len(left_index):
            # Compute distances from the segment start point to the all
            # the leftover points
            dists = sm.distance.cdist(
                np.atleast_2d(seg_start), coordinates[left_index]
            ).flatten()

            # Select the points that belong to this segment
            dist_mask = dists <= segment_length
            pass_index = np.where(dist_mask)[0]
            fail_index = np.where(~dist_mask)[0]
            seg_index = left_index[pass_index]

            # If the next closest point is backwards, make it the last segment
            if len(fail_index):
                next_id = np.argmin(dists[fail_index])
                next_index = left_index[fail_index][next_id]
                next_closest = coordinates[next_index]
                if np.dot(next_closest - seg_start, seg_start - start_point) < 0.0:
                    fail_index = fail_index[:0]
                    if not len(pass_index):
                        break

            # Estimate the direction of the segment
            seg_coords = coordinates[seg_index]
            if (
                method == "step"
                and len(seg_index) > min_count
                and np.max(dists[pass_index]) > 0.0
            ):
                # Estimate direction w.r.t. the segment start point ('step')
                direction = sm.mean(seg_coords - seg_start, axis=0)
            elif len(fail_index):
                # Take direction as the vector joining the next closest point
                # ('step_next'). Also apply this method is the `min_count`
                # condition is not satisfied for this segment.
                direction = next_closest - seg_start
            else:
                # If this is the last segment, find the farthest point from
                # its start and make it the last point.
                last_id = np.argmax(dists[pass_index])
                direction = seg_coords[last_id] - seg_start

            if np.linalg.norm(direction):
                direction /= np.linalg.norm(direction)
            else:
                direction = np.array([1.0, 0.0, 0.0], dtype=coordinates.dtype)

            # Compute the segment length. If it's the last segment,
            # track the distance to the fathest point in the segment
            if len(fail_index):
                length = segment_length
            else:
                last_id = np.argmax(dists[pass_index])
                to_end = seg_coords[last_id] - seg_start
                length = np.dot(to_end, direction)

            # Step the segment start point in the direction of the segment
            seg_start += length * direction

            # Update the leftover index
            left_index = left_index[fail_index]

            # Store the segment information
            seg_clusts.append(seg_index)
            seg_dirs_l.append(direction)
            seg_lengths_l.append(length)

        # Convert lists of directions and lengths to numpy.ndarray objects
        shape = (len(seg_dirs_l), coordinates.shape[1])
        seg_dirs = np.empty(shape, dtype=coordinates.dtype)
        seg_lengths = np.empty(shape[0], dtype=coordinates.dtype)
        for i in range(len(seg_clusts)):
            seg_dirs[i] = seg_dirs_l[i]
            seg_lengths[i] = seg_lengths_l[i]

        return seg_clusts, seg_dirs, seg_lengths

    elif method == "bin_pca":
        # Find the principal component of the whole track
        track_dir = sm.decomposition.principal_components(coordinates)[0]

        # If a track end point is provided, check which end the track end point
        # is on and flip the principal axis coordinates, if needed
        pcoordinates = np.dot(coordinates, track_dir)
        if point is not None:
            pstart = np.dot(point, track_dir)
            if np.abs(np.min(pcoordinates) - pstart) > np.abs(
                np.max(pcoordinates) - pstart
            ):
                pstart = -pstart
                pcoordinates = -pcoordinates

        # Bin the track along the principal component vector
        if point is not None:
            boundaries = np.arange(
                min(pstart, np.min(pcoordinates)), np.max(pcoordinates), segment_length
            )
        else:
            boundaries = np.arange(
                np.min(pcoordinates), np.max(pcoordinates), segment_length
            )

        seg_labels = np.digitize(pcoordinates, boundaries)
        seg_clusts = nb.typed.List.empty_list(np.empty(0, dtype=np.int64))
        for l in range(len(boundaries)):
            seg_clusts.append(np.where(seg_labels == l + 1)[0])

        # Compute the segment directions and lengths
        seg_dirs = np.empty(
            (len(seg_clusts), coordinates.shape[1]), dtype=coordinates.dtype
        )
        seg_lengths = np.empty(len(seg_clusts), dtype=coordinates.dtype)
        for i, seg in enumerate(seg_clusts):
            # If this segment is empty, use track-level information
            if not len(seg):
                seg_dirs[i] = track_dir
                seg_lengths[i] = segment_length
                continue

            # Compute principal component of the segment, use it as direction
            if len(seg) > min_count:
                direction = sm.decomposition.principal_components(coordinates[seg])[0]
                if np.dot(direction, track_dir) < 0.0:
                    direction = -direction
            else:
                direction = track_dir
            seg_dirs[i] = direction

            # Evaluate length of the segment as constrained by the track
            # principal axis bin that defines it
            if i < len(seg_clusts) - 1:
                length = segment_length
            else:
                length = np.max(pcoordinates) - boundaries[-1]

            costh = np.dot(direction, track_dir)
            if costh != 0.0:
                length /= costh

            seg_lengths[i] = length

        return seg_clusts, seg_dirs, seg_lengths

    else:
        raise ValueError("Track segmentation method not recognized")


def get_track_spline(coordinates, segment_length, s=None):
    """Estimate the best approximating curve defined by a point cloud using
    univariate 3D splines.

    The length is computed by measuring the length of the piecewise linear
    interpolation of the spline at points defined by the bin size.

    Parameters
    ----------
    coordinatea : np.ndarray
        (N, 3) point cloud
    segment_length : float
        The subdivision length at which to sample points from the spline.
        If the track length is less than the segment_length, then the returned
        length will be computed from the farthest two projected points along
        the track's principal direction.
    s : float, optional
        The smoothing factor to be used in spline regression, by default None

    Returns
    -------
    u : np.ndarray
        (N) The principal axis parametrization of the curve
        C(u) = (spx(u), spy(u), spz(u))
    sppoints : np.ndarray
        (N, 3) The graph of the spline at points u
    splines : scipy.interpolate.UnivariateSpline
        Approximating splines for the point cloud defined by points
    length : float
        The estimate of the total length of the curve
    """
    # Compute the principal component along which to segment the track
    track_dir = sm.decomposition.principal_components(coordinates)[0]
    pcoords = np.dot(coordinates, track_dir)
    perm = np.argsort(pcoords.squeeze())
    u = pcoords[perm]

    # If there is less than four points, cannot fit a 3D spline
    if len(coordinates) < 4:
        # Fall back on displacement to estimate length
        length = np.max(pcoords) - np.min(pcoords)
        return u.squeeze, None, None, length

    # Compute the univariate splines along each axis
    spx = UnivariateSpline(u, coordinates[perm][:, 0], s=s)
    spy = UnivariateSpline(u, coordinates[perm][:, 1], s=s)
    spz = UnivariateSpline(u, coordinates[perm][:, 2], s=s)
    sppoints = np.hstack([spx(u), spy(u), spz(u)])
    splines = [spx, spy, spz]

    # If track length is less than segment_length, just return length.
    # Otherwise estimate length by piecewise linear interpolation
    start, end = u.min(), u.max()
    length = end - start
    if length > segment_length:
        bins = np.arange(u.min(), u.max(), segment_length)
        bins = np.hstack([bins, np.array([u.max()])])
        pt_approx = np.hstack([sp(bins).reshape(-1, 1) for sp in splines])
        segments = np.linalg.norm(pt_approx[1:] - pt_approx[:-1], axis=1)
        length = segments.sum()

    return u.squeeze(), sppoints, splines, length
